"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

"""
Torch batch sampler - balancing per batch
"""
import logging
import math
from typing import Any, List, Optional

import numpy as np
from torch.utils.data.sampler import Sampler

from fuse.data.dataset.dataset_base import FuseDatasetBase
from fuse.utils.utils_debug import FuseUtilsDebug
from fuse.utils.utils_logger import log_object_input_state


class FuseSamplerBalancedBatch(Sampler):
    """
    Torch batch sampler - balancing per batch
    """

    def __init__(self, dataset: FuseDatasetBase, balanced_class_name: str, num_balanced_classes: int, batch_size: int,
                 balanced_class_weights: Optional[List[int]] = None, balanced_class_probs: Optional[List[float]] = None,
                 num_batches: Optional[int] = None, use_dataset_cache: bool = False) -> None:
        """
        :param dataset: dataset used to extract the balanced class from each sample
        :param balanced_class_name:  the name of balanced class to extract from dataset
        :param num_balanced_classes: number of classes to balance between
        :param batch_size: batch_size.
                        - If balanced_class_weights=Nobe, Must be divided by num_balanced_classes
                        - Otherwise must be equal to sum of balanced_class_weights
        :param balanced_class_weights: Optional, integer per balanced class,
                                        specifying the number of samples from each class to include in each batch.
                                        If not specified and equal number of samples from each class will be used.
        :param balanced_class_probs: Optional, probability per class. Random sampling approach will be performed.
                                     such that an epoch will go over all the data at least once.
        :param num_batches: Optional, Set number of batches. If not set. The number of batches will automatically set.

        :param use_dataset_cache: to retrieve the balanced class from dataset try to use caching.
                                 Should be set to True if reading it from cache is faster than running the single processor
        """
        # log object
        log_object_input_state(self, locals())

        super().__init__(None)

        # store input
        self.dataset = dataset
        self.balanced_class_name = balanced_class_name
        self.num_balanced_classes = num_balanced_classes
        self.batch_size = batch_size
        self.balanced_class_weights = balanced_class_weights
        self.balanced_class_probs = balanced_class_probs
        self.num_batches = num_batches
        self.use_dataset_cache = use_dataset_cache

        # validate input
        if balanced_class_weights is not None and balanced_class_probs is not None:
            raise Exception('Set either balanced_class_weights or balanced_class_probs, not both.')
        elif balanced_class_weights is None and balanced_class_probs is None:
            if batch_size % num_balanced_classes != 0:
                raise Exception(f'batch_size ({batch_size}) % num_balanced_classes ({num_balanced_classes}) must be 0')
        elif balanced_class_weights is not None:
            if len(balanced_class_weights) != num_balanced_classes:
                raise Exception(
                    f'Expecting balance_class_weights ({balanced_class_weights}) to have a weight per balanced class ({num_balanced_classes})')
            if sum(balanced_class_weights) != batch_size:
                raise Exception(f'balanced_class_weights {balanced_class_weights} expected to sum up to batch_size {batch_size}')
        else:
            # noinspection PyTypeChecker
            if len(balanced_class_probs) != num_balanced_classes:
                raise Exception(
                    f'Expecting balance_class_probs ({balanced_class_probs}) to have a probability per balanced class ({num_balanced_classes})')
            if not math.isclose(sum(balanced_class_probs), 1.0):
                raise Exception(f'balanced_class_probs {balanced_class_probs} expected to sum up to 1.0')

        # if weights not specified, set weights to equally balance per batch
        if self.balanced_class_weights is None and self.balanced_class_probs is None:
            self.balanced_class_weights = [self.batch_size // self.num_balanced_classes] * self.num_balanced_classes

        lgr = logging.getLogger('Fuse')
        lgr.debug(f'FuseSamplerBalancedBatch: balancing per batch - balanced_class_name {self.balanced_class_name}, '
                  f'batch_size={batch_size}, weights={self.balanced_class_weights}, probs={self.balanced_class_probs}')

        # get balanced classes per each sample
        self.balanced_classes = dataset.get(None, self.balanced_class_name, use_cache=use_dataset_cache)
        self.balanced_classes = np.array(self.balanced_classes)
        self.balanced_class_indices = [np.where(self.balanced_classes == cls_i)[0] for cls_i in range(self.num_balanced_classes)]
        self.balanced_class_sizes = [len(self.balanced_class_indices[cls_i]) for cls_i in range(self.num_balanced_classes)]
        lgr.debug('FuseSamplerBalancedBatch: samples per each balanced class {}'.format(self.balanced_class_sizes))

        # debug - simple batch
        batch_mode = FuseUtilsDebug().get_setting('sampler_batch_mode')
        if batch_mode == 'simple':
            num_avail_bcls = 0
            for bcls_num_samples in self.balanced_class_sizes:
                if bcls_num_samples != 0:
                    num_avail_bcls += 1

            self.balanced_class_weights = None
            self.balanced_class_probs = [1.0/num_avail_bcls if bcls_num_samples != 0 else 0.0 for bcls_num_samples in self.balanced_class_sizes]
            lgr.info('FuseSamplerBalancedBatch: debug mode - override to random sample')

        # calc batch index to balanced class mapping according to weights
        if self.balanced_class_weights is not None:
            self.batch_index_to_class = []
            for balanced_cls in range(self.num_balanced_classes):
                self.batch_index_to_class.extend([balanced_cls] * self.balanced_class_weights[balanced_cls])
        else:
            # probabilistic method - will be randomly select per epoch
            self.batch_index_to_class = None

        # make sure that size != 0 for all balanced classes
        for cls_size in enumerate(self.balanced_class_sizes):
            if (self.balanced_class_weights is not None and self.balanced_class_weights != 0) or (
                    self.balanced_class_probs is not None and self.balanced_class_probs != 0.0):
                if cls_size == 0:
                    msg = f'Every balanced class must include at least one sample (num of samples per balanced class{self.balanced_class_sizes})'
                    raise Exception(msg)

        # Shuffle balanced class indices
        for indices in self.balanced_class_indices:
            np.random.shuffle(indices)

        # Calculate num batches. Number of batches to iterate over all data at least once
        # Calculate only if not directly specified by the user
        if self.num_batches is None:
            if self.balanced_class_weights is not None:
                balanced_class_weighted_sizes = [self.balanced_class_sizes[cls_i] // self.balanced_class_weights[cls_i] if self.balanced_class_weights[cls_i] != 0 else 0 for cls_i in
                                                 range(self.num_balanced_classes)]
            else:
                # approximate size!
                balanced_class_weighted_sizes = [
                    self.balanced_class_sizes[cls_i] // (self.balanced_class_probs[cls_i] * self.batch_size) if self.balanced_class_probs[
                                                                                                                    cls_i] != 0.0 else 0 for
                    cls_i in range(self.num_balanced_classes)]
            bigger_balanced_class_weighted_size = max(balanced_class_weighted_sizes)
            self.num_batches = int(bigger_balanced_class_weighted_size) + 1
            lgr.debug(f'FuseSamplerBalancedBatch: num_batches = {self.num_batches}')

        # pointers per class
        self.cls_pointers = [0] * self.num_balanced_classes
        self.sample_pointer = 0

    def __iter__(self) -> np.ndarray:
        for batch_idx in range(self.num_batches):
            yield self._make_batch()

    def __len__(self) -> int:
        return self.num_batches

    def _get_sample(self, balanced_class: int) -> Any:
        """
        sample index given balanced class value
        :param balanced_class: integer representing balanced class value
        :return: sample index
        """
        if self.balanced_class_indices[balanced_class].shape[0] == 0:
            msg = f'There are no samples in balanced class {balanced_class}'
            logging.getLogger('Fuse').error(msg)
            raise Exception(msg)

        sample_idx = self.balanced_class_indices[balanced_class][self.cls_pointers[balanced_class]]

        self.cls_pointers[balanced_class] += 1
        if self.cls_pointers[balanced_class] == self.balanced_class_sizes[balanced_class]:
            self.cls_pointers[balanced_class] = 0
            np.random.shuffle(self.balanced_class_indices[balanced_class])

        return sample_idx

    def _make_batch(self) -> list:
        """
        :return: list of indices to collate batch
        """
        if self.batch_index_to_class is not None:
            batch_index_to_class = self.batch_index_to_class
        else:
            # calc one according to probabilities
            batch_index_to_class = np.random.choice(np.arange(self.num_balanced_classes), self.batch_size, p=self.balanced_class_probs)
        batch_sample_indices = []
        for batch_index in range(self.batch_size):
            balanced_class = batch_index_to_class[batch_index]
            batch_sample_indices.append(self._get_sample(balanced_class))

        np.random.shuffle(batch_sample_indices)
        return batch_sample_indices



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

import logging
import traceback
from typing import Dict, Optional, Callable, List

import numpy as np
import sklearn.metrics

from fuse.metrics.metric_base import FuseMetricBase
from fuse.metrics.metrics_toolbox import FuseMetricsToolBox


class FuseMetricROCCurve(FuseMetricBase):
    """
    Output the roc curve (Multi class version - one vs rest)
    """

    def __init__(self,
                 pred_name: str,
                 target_name: str,
                 output_filename: str,
                 class_names: Optional[List[str]] = None,
                 filter_func: Optional[Callable] = None,
                 use_sample_weights: Optional[bool] = False,
                 sample_weight_name: Optional[str] = None) -> None:
        """
        :param pred_name:           batch_dict key for predicted output (e.g., class probabilities after softmax)
        :param target_name:         batch_dict key for target (e.g., ground truth label)
        :param output_filename:     output filename to save the figure
        :param class_names:         Optional - name for each class otherwise the class index will be used.
        :param filter_func:         Optional - function that filters batch_dict. The function gets as input batch_dict and returns filtered batch_dict.
                                    When None (default), the entire batch dict is collected and processed.
        :param use_sample_weights:  Optional - when True (default is False), metrics are computed with sample weigths.
                                    For this, use argument 'sample_weight_name' to define the batch_dict key for the sample weight
        :param sample_weight_name:  Optional, when use_sample_weights is True, user must specify the key for collecting sample weights
        """
        super().__init__(pred_name, target_name, filter_func=filter_func,
                         use_sample_weights=use_sample_weights, sample_weight_name=sample_weight_name)
        self._class_names = class_names
        self._output_filename = output_filename

    def process(self) -> Dict[str, float]:
        """
        Save to a file ROC curve and AUC values.
        :return: an empty dictionary
        """
        results = {}
        num_classes = self.collected_data['pred'][0].shape[0]
        class_names = FuseMetricsToolBox.get_class_names(num_classes, self._class_names)

        sample_weight = self.collected_data['sample_weight'] if self.use_sample_weights else None

        try:
            FuseMetricsToolBox.roc_curve(self.epoch_preds, self.epoch_targets, sample_weight, class_names, self._output_filename)
        except:
            track = traceback.format_exc()
            logging.getLogger('Fuse').error(track)

        return results




"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

"""
Augmentor Base class
"""
from abc import ABC, abstractmethod
from typing import Any


class FuseAugmentorBase(ABC):
    """
    Base class for augmentor.
    Given an augmenatation pipline description, expected to sample random parameters first and then apply them.
    """

    @abstractmethod
    def get_random_augmentation_desc(self) -> Any:
        """
        Sample random parameters for augmentation
        :return:
        """
        raise NotImplementedError

    @abstractmethod
    def apply_augmentation(self, sample: Any, augmentation_desc: Any) -> Any:
        """
        Apply the augmenation according to the given parameters. Must be deterministic.
        :param sample: the original sample as generated by the dataset
        :param augmentation_desc: augmentation parameters. Output of get_random_augmentation_desc()
        :return: augmented sample
        """
        raise NotImplementedError

    @abstractmethod
    def summary(self) -> str:
        """
        String summary of the object
        """
        raise NotImplementedError

    def __call__(self, sample: Any):
        """
        generate random and apply the augmentation at once.
        :param sample:
        :return:
        """
        augmentation_desc = self.get_random_augmentation_desc()
        sample_aug = self.apply_augmentation(sample, augmentation_desc)

        return sample_aug



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

import logging
import traceback
from typing import Dict, Optional, Callable, List

import numpy as np
import sklearn.metrics

from Fuse.metrics.metric_base import FuseMetricBase


class FuseMetricPartialAUC(FuseMetricBase):
    """
    Partial area under receiver operating characteristic curve
    Multi class version
    Will add to epoch results:
    metrics.<metric_name>.<cls> - partial auc per class vs rest - where <cls> will be the class name if supplied or class index
    metric.<metric_name>.macro_avg - average of all class vs rest results
    """

    def __init__(self,
                 pred_name: str,
                 target_name: str,
                 tpr_val: float,
                 class_names: Optional[List[str]] = None,
                 filter_func: Optional[Callable] = None,
                 use_sample_weights: Optional[bool] = False,
                 sample_weight_name: Optional[str] = None) -> None:
        """
        :param pred_name:           batch_dict key for predicted output (e.g., class probabilities after softmax)
        :param target_name:         batch_dict key for target (e.g., ground truth label)
        :param tpr_val:             the minimum value of sensitivity to compute partial AUC
        :param class_names:         Optional - name for each class otherwise the class index will be used.
        :param filter_func:         Optional - function that filters batch_dict. The function gets as input batch_dict and returns filtered batch_dict.
                                    When None (default), the entire batch dict is collected and processed.
        :param use_sample_weights:  Optional - when True (default is False), metrics are computed with sample weigths.
                                    For this, use argument 'sample_weight_name' to define the batch_dict key for the sample weight
        :param sample_weight_name:  Optional, when use_sample_weights is True, user must specify the key for collecting sample weights
        """
        super().__init__(pred_name, target_name, filter_func=filter_func,
                         use_sample_weights=use_sample_weights, sample_weight_name=sample_weight_name)
        self._class_names = class_names
        self._tpr_val = tpr_val

    def process(self) -> Dict[str, float]:
        """
        Calculates partial ROCAUC per epoch
        :return: dictionary including Partial Area under ROC curve (floating point in range [0, 1]/-1) for each class (class vs rest),
                 -1 will be set for invalid/undefined result (an error message will be printed)
                 The dictionary will also include the average Partial AUC
        """
        results = {}
        paucs_per_class = []
        if self._class_names is None:
            num_classes = self.collected_data['pred'][0].shape[0]
            class_names = ['class_' + str(i) for i in range(num_classes)]
        else:
            class_names = self._class_names
        sample_weight = self.collected_data['sample_weight'] if self.use_sample_weights else None
        try:
            for cls_ind, cls in enumerate(class_names):
                predicted_ind = [pred[cls_ind] for pred in self.collected_data['pred']]
                fpr, tpr, _ = sklearn.metrics.roc_curve(self.collected_data['target'], predicted_ind, pos_label=cls_ind, sample_weight=sample_weight)
                truncated_tprs = np.maximum(0, tpr - self._tpr_val)
                pauc = sklearn.metrics.auc(fpr, truncated_tprs)
                pauc = pauc / (1 - self._tpr_val)  # normalize
                results[cls] = pauc
                paucs_per_class.append(pauc)

            results['macro_avg'] = np.nanmean(paucs_per_class)
        except:
            track = traceback.format_exc()
            logging.getLogger('Fuse').error(track)
            results = {cls: -1 for cls in class_names}
            results['macro_avg'] = -1
        return results


if __name__ == '__main__':
    data = {'preds': np.array([[0.8, 0.1, 0.1],
                               [0.5, 0.3, 0.2],
                               [0.6, 0.3, 0.1],
                               [0.6, 0.1, 0.3],
                               [0.7, 0.2, 0.1],
                               [0.3, 0.2, 0.5],
                               [0.1, 0.2, 0.7],
                               [0.2, 0.6, 0.2],
                               [0.3, 0.3, 0.4],
                               [0.7, 0.2, 0.1]]),
            'targets': np.array([0, 0, 0, 0, 0, 1, 1, 1, 2, 2]),
            'weights': np.array([0.03, 0.9, 0.05, 0.52, 0.23, 0.72, 0.13, 0.113, 0.84, 0.09])}

    metric = FuseMetricPartialAUC(pred_name='preds', target_name='targets', tpr_val=0.95, use_sample_weights=True,
                           sample_weight_name='weights')
    metric.collect(data)
    res = metric.process()
    print("use_sample_weights = True")
    for k, v in res.items():
        print(k, '\t', v)

    metric = FuseMetricPartialAUC(pred_name='preds', target_name='targets', tpr_val=0.95)
    metric.collect(data)
    res = metric.process()
    print("use_sample_weights = False")
    for k, v in res.items():
        print(k, '\t', v)


"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

import os
import requests
import zipfile
import io
import wget

import logging


def download_and_extract_isic(root_data: str, year: str = '2016'):
    """
    Download images and metadata from ISIC challenge
    :param root_data: path where data should be located
    :param year: ISIC challenge year (2016 or 2017)
    """
    lgr = logging.getLogger('Fuse')

    if year == '2016':
        # 2016 - Train
        if not os.path.exists(os.path.join(root_data, 'data/ISIC2016_Training_Data')):
            lgr.info('Extract ISIC-2016 training data...')

            url = 'https://isic-challenge-data.s3.amazonaws.com/2016/ISBI2016_ISIC_Part3_Training_Data.zip'
            r = requests.get(url)
            z = zipfile.ZipFile(io.BytesIO(r.content))
            z.extractall(os.path.join(root_data, 'data'))
            os.rename(os.path.join(root_data, 'data/ISBI2016_ISIC_Part3_Training_Data'),
                      os.path.join(root_data, 'data/ISIC2016_Training_Data'))

            lgr.info('Extracting ISIC-2016 training data: done')

        if not os.path.exists(os.path.join(root_data, 'data/ISIC2016_Training_GroundTruth.csv')):
            url = 'https://isic-challenge-data.s3.amazonaws.com/2016/ISBI2016_ISIC_Part3_Training_GroundTruth.csv'
            wget.download(url, os.path.join(root_data, 'data/ISIC2016_Training_GroundTruth.csv'))

        # 2016 - Test
        if not os.path.exists(os.path.join(root_data, 'data/ISIC2016_Test_Data')):
            lgr.info('Extract ISIC-2016 test data...')

            url = 'https://isic-challenge-data.s3.amazonaws.com/2016/ISBI2016_ISIC_Part3_Test_Data.zip'
            r = requests.get(url)
            z = zipfile.ZipFile(io.BytesIO(r.content))
            z.extractall(os.path.join(root_data, 'data'))
            os.rename(os.path.join(root_data, 'data/ISBI2016_ISIC_Part3_Test_Data'),
                      os.path.join(root_data, 'data/ISIC2016_Test_Data'))

            lgr.info('Extracting ISIC-2016 test data: done')

        if not os.path.exists(os.path.join(root_data, 'data/ISIC2016_Test_GroundTruth.csv')):
            url = 'https://isic-challenge-data.s3.amazonaws.com/2016/ISBI2016_ISIC_Part3_Test_GroundTruth.csv'
            wget.download(url, os.path.join(root_data, 'data/ISIC2016_Test_GroundTruth.csv'))

    if year == '2017':
        # 2017 - Train
        if not os.path.exists(os.path.join(root_data, 'data/ISIC2017_Training_Data')):
            lgr.info('\nExtract ISIC-2017 training data... (this may take a few minutes)')

            url = 'https://isic-challenge-data.s3.amazonaws.com/2017/ISIC-2017_Training_Data.zip'
            r = requests.get(url)
            z = zipfile.ZipFile(io.BytesIO(r.content))
            z.extractall(os.path.join(root_data, 'data'))
            os.rename(os.path.join(root_data, 'data/ISIC-2017_Training_Data'),
                      os.path.join(root_data, 'data/ISIC2017_Training_Data'))

            lgr.info('Extracting ISIC-2017 training data: done')

        if not os.path.exists(os.path.join(root_data, 'data/ISIC2017_Training_GroundTruth.csv')):
            url = 'https://isic-challenge-data.s3.amazonaws.com/2017/ISIC-2017_Training_Part3_GroundTruth.csv'
            wget.download(url, os.path.join(root_data, 'data/ISIC2017_Training_GroundTruth.csv'))

        # 2017 - Validation
        if not os.path.exists(os.path.join(root_data, 'data/ISIC2017_Validation_Data')):
            lgr.info('\nExtract ISIC-2017 validation data... (this may take a few minutes)')

            url = 'https://isic-challenge-data.s3.amazonaws.com/2017/ISIC-2017_Validation_Data.zip'
            r = requests.get(url)
            z = zipfile.ZipFile(io.BytesIO(r.content))
            z.extractall(os.path.join(root_data, 'data'))
            os.rename(os.path.join(root_data, 'data/ISIC-2017_Validation_Data'),
                      os.path.join(root_data, 'data/ISIC2017_Validation_Data'))

            lgr.info('Extracting ISIC-2017 validation data: done')

        if not os.path.exists(os.path.join(root_data, 'data/ISIC2017_Validation_GroundTruth.csv')):
            url = 'https://isic-challenge-data.s3.amazonaws.com/2017/ISIC-2017_Validation_Part3_GroundTruth.csv'
            wget.download(url, os.path.join(root_data, 'data/ISIC2017_Validation_GroundTruth.csv'))

        # 2017 - Test
        if not os.path.exists(os.path.join(root_data, 'data/ISIC2017_Test_Data')):
            lgr.info('\nExtracting ISIC-2017 test data... (this may take a few minutes)')

            url = 'https://isic-challenge-data.s3.amazonaws.com/2017/ISIC-2017_Test_v2_Data.zip'
            r = requests.get(url)
            z = zipfile.ZipFile(io.BytesIO(r.content))
            z.extractall(os.path.join(root_data, 'data'))
            os.rename(os.path.join(root_data, 'data/ISIC-2017_Test_v2_Data'),
                      os.path.join(root_data, 'data/ISIC2017_Test_Data'))

            lgr.info('Extracting ISIC-2017 test data: done')

        if not os.path.exists(os.path.join(root_data, 'data/ISIC2017_Test_GroundTruth.csv')):
            url = 'https://isic-challenge-data.s3.amazonaws.com/2017/ISIC-2017_Test_v2_Part3_GroundTruth.csv'
            wget.download(url, os.path.join(root_data, 'data/ISIC2017_Test_GroundTruth.csv'))






"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

from typing import Callable, Dict, Optional

import torch

from fuse.losses.loss_base import FuseLossBase
from fuse.utils.utils_hierarchical_dict import FuseUtilsHierarchicalDict


class FuseLossDefault(FuseLossBase):
    """
    Default Fuse loss function
    """

    def __init__(self,
                 pred_name: str = None,
                 target_name: str = None,
                 batch_kwargs_name: str = None,
                 callable: Callable = None,
                 sample_weight_name: Optional[str] = None,
                 weight: float = 1.0,
                 filter_func: Optional[Callable] = None,
                 **kwargs
                 ) -> None:
        """
        This class wraps a PyTorch loss function with a Fuse api.
        :param pred_name:               batch_dict key for prediction (e.g., network output)
        :param target_name:             batch_dict key for target (e.g., ground truth label)
        :param batch_kwargs_name:       batch_dict key for additional, ad-hoc kwargs for loss function
                                        Note: batch_kwargs will be merged into other loss function kwargs
        :param sample_weight_name       batch_dict key that holds the sample weight for loss summation
        :param callable:                PyTorch loss function handle (e.g., torch.nn.functional.cross_entropy)

        :param weight:                  Weight multiplier for final loss value
        :param filter_func:             function that filters batch_dict/ The function gets ans input batch_dict and returns filtered batch_dict
        :param kwargs:                  kwargs for PyTorch loss function
        """
        super().__init__()
        self.pred_name = pred_name
        self.target_name = target_name
        self.batch_kwargs_name = batch_kwargs_name
        self.callable = callable
        self.sample_weight_name = sample_weight_name
        self.weight = weight
        self.filter_func = filter_func
        self.kwargs = kwargs

    def __call__(self, batch_dict: Dict) -> torch.Tensor:
        # filter batch_dict if required
        if self.filter_func is not None:
            batch_dict = self.filter_func(batch_dict)
        preds = FuseUtilsHierarchicalDict.get(batch_dict, self.pred_name)
        targets = FuseUtilsHierarchicalDict.get(batch_dict, self.target_name)
        batch_kwargs = FuseUtilsHierarchicalDict.get(batch_dict, self.batch_kwargs_name) if self.batch_kwargs_name is not None else {}
        kwargs_copy = self.kwargs.copy()
        kwargs_copy.update(batch_kwargs)
        if self.sample_weight_name is not None:
            assert 'reduction' not in kwargs_copy.keys(), 'reduction is forced to none when applying sample weight'
            kwargs_copy.update({'reduction': 'none'})
        loss_obj = self.callable(preds, targets, **kwargs_copy) * self.weight
        if self.sample_weight_name is not None:
            sample_weight = FuseUtilsHierarchicalDict.get(batch_dict, self.sample_weight_name)
            weighted_loss = loss_obj*sample_weight
            loss_obj = torch.mean(weighted_loss)

        return loss_obj


if __name__ == '__main__':
    import torch

    batch_dict = {'pred': torch.randn(3, 5, requires_grad=True),
                  'gt': torch.empty(3, dtype=torch.long).random_(5),
                  'batch_loss_kwargs': {'reduction': 'mean', 'ignore_index': 0}}

    loss = FuseLossDefault(pred_name='pred',
                           target_name='gt',
                           batch_kwargs_name='batch_loss_kwargs',
                           callable=torch.nn.functional.cross_entropy,
                           weight=1.0,
                           reduction='sum')

    res = loss(batch_dict)
    print('Loss output = ' + str(res))



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

# slightly modified (error message only) version of https://github.com/davda54/sam/blob/main/sam.py distributed under MIT license
#
# Copyright ©
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
from typing import Dict

import torch

from fuse.utils.utils_hierarchical_dict import FuseUtilsHierarchicalDict
from fuse.managers.callbacks.callback_base import FuseCallback
from fuse.managers.manager_state import FuseManagerState


class SAM(torch.optim.Optimizer):
    """
    SAM optimizer - see https://github.com/davda54/sam/blob/main/sam.py
    To use in FuseMedML:
    Create the optimizer and add FuseCallbackSamOpt() to the list of callbacks:.
    Examples:
    base_optimizer = torch.optim.SGD
    optimizer = SAM(model.parameters(), base_optimizer,
                    lr=lr,
                    momentum=momentum,
                    weight_decay=weight_decay)
    callbacks.append(FuseCallbackSamOpt())

    """
    def __init__(self, params, base_optimizer, rho=0.05, **kwargs):
        assert rho >= 0.0, f"Invalid rho, should be non-negative: {rho}"

        defaults = dict(rho=rho, **kwargs)
        super(SAM, self).__init__(params, defaults)

        self.base_optimizer = base_optimizer(self.param_groups, **kwargs)
        self.param_groups = self.base_optimizer.param_groups

    @torch.no_grad()
    def first_step(self, zero_grad=False):
        grad_norm = self._grad_norm()
        for group in self.param_groups:
            scale = group["rho"] / (grad_norm + 1e-12)

            for p in group["params"]:
                if p.grad is None: continue
                e_w = p.grad * scale.to(p)
                p.add_(e_w)  # climb to the local maximum "w + e(w)"
                self.state[p]["e_w"] = e_w

        if zero_grad: self.zero_grad()

    @torch.no_grad()
    def second_step(self, zero_grad=False):
        for group in self.param_groups:
            for p in group["params"]:
                if p.grad is None: continue
                p.sub_(self.state[p]["e_w"])  # get back to "w" from "w + e(w)"

        self.base_optimizer.step()  # do the actual "sharpness-aware" update

        if zero_grad: self.zero_grad()

    def step(self, closure=None):
        # do nothing, instead use first_step and second_step
        return

    def _grad_norm(self):
        shared_device = self.param_groups[0]["params"][0].device  # put everything on the same device, in case of model parallelism
        norm = torch.norm(
            torch.stack([
                p.grad.norm(p=2).to(shared_device)
                for group in self.param_groups for p in group["params"]
                if p.grad is not None
            ]),
            p=2
        )
        return norm


class FuseCallbackSamOpt(FuseCallback):
    """
    Use this callback for SAM optimizer
    """

    def on_train_begin(self, state: FuseManagerState):
        self.state = state

    def on_batch_end(self, mode: str, batch: int, batch_dict: Dict = None):
        # self.virtual_batch.append(batch_dict)
        if mode == 'train':
            self.state.optimizer.first_step(zero_grad=True)
            # forward net
            batch_dict['model'] = self.state.net(batch_dict)
            # compute total loss and keep loss results
            total_loss: torch.Tensor = 0
            for loss_name, loss_function in self.state.losses.items():
                current_loss_result = loss_function(batch_dict)
                FuseUtilsHierarchicalDict.set(batch_dict, 'losses.' + loss_name, current_loss_result.data.item())
                # sum all losses for backward
                total_loss += current_loss_result
            total_loss.backward()
            self.state.optimizer.second_step(zero_grad=True)



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

from typing import Dict, List, Sequence

from fuse.data.augmentor.augmentor_default import FuseAugmentorDefault
from fuse.managers.callbacks.callback_base import FuseCallback


class FuseAugmentorBatchCallback(FuseCallback):
    """
    Simple class which gets augmentation pipeline and apply augmentation on a batch level batch dict
    """
    def __init__(self, aug_pipeline: List, modes: Sequence[str] = ('train',)):
        """
        :param aug_pipeline: See  FuseAugmentorDefault
        :param modes: modees to apply the augmentation: 'train', 'validation' and/or 'infer'
        """
        self._augmentor = FuseAugmentorDefault(aug_pipeline)
        self._modes = modes

    def on_data_fetch_end(self, mode: str, batch: int, batch_dict: Dict = None) -> None:
        if mode in self._modes:
            self._augmentor(batch_dict)


"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

import unittest
import pandas as pd
import torch

from fuse.data.processor.processor_dataframe import FuseProcessorDataFrame


class FuseProcessorDataFrameTestCase(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def setUp(self):
        pass

    def test_all_columns(self):
        df = pd.DataFrame({"desc": ['one', 'two', 'three'], "int_val": [4, 5, 6], "string_val": ['val1', 'val2', 'val3']})
        proc = FuseProcessorDataFrame(data=df, sample_desc_column='desc')

        self.assertDictEqual(proc('one'), {'int_val': 4, 'string_val':'val1'})
        self.assertDictEqual(proc('two'), {'int_val': 5, 'string_val':'val2'})
        self.assertDictEqual(proc('three'), {'int_val': 6, 'string_val':'val3'})

    def test_rename(self):
        df = pd.DataFrame({"desc": ['one', 'two', 'three'], "int_val": [4, 5, 6], "string_val": ['val1', 'val2', 'val3']})
        proc = FuseProcessorDataFrame(data=df, sample_desc_column='desc', rename_columns={'int_val': 'new_name'})

        self.assertDictEqual(proc('one'), {'new_name': 4, 'string_val':'val1'})
        self.assertDictEqual(proc('two'), {'new_name': 5, 'string_val':'val2'})
        self.assertDictEqual(proc('three'), {'new_name': 6, 'string_val':'val3'})

    def test_specific_columns(self):
        df = pd.DataFrame({"desc": ['one', 'two', 'three'], "int_val": [4, 5, 6], "string_val": ['val1', 'val2', 'val3']})
        proc = FuseProcessorDataFrame(data=df, sample_desc_column='desc', rename_columns={'int_val': 'new_name'},
                                      columns_to_extract=['string_val', 'desc'])

        self.assertDictEqual(proc('one'), {'string_val':'val1'})
        self.assertDictEqual(proc('two'), {'string_val':'val2'})
        self.assertDictEqual(proc('three'), {'string_val':'val3'})

    def test_tensors(self):
        df = pd.DataFrame({"desc": ['one', 'two', 'three'], "int_val": [4, 5, 6], "string_val": ['val1', 'val2', 'val3']})
        proc = FuseProcessorDataFrame(data=df, sample_desc_column='desc', columns_to_tensor=['int_val', 'invalid_column'])

        self.assertDictEqual(proc('one'), {'int_val': torch.tensor(4, dtype=torch.int64), 'string_val':'val1'})
        self.assertDictEqual(proc('two'), {'int_val': torch.tensor(5, dtype=torch.int64), 'string_val':'val2'})
        self.assertDictEqual(proc('three'), {'int_val': torch.tensor(6, dtype=torch.int64), 'string_val':'val3'})

    def test_tensors_with_types(self):
        df = pd.DataFrame({"desc": ['one', 'two', 'three'], "int_val": [4, 5, 6], "float_val": [4.1, 5.3, 6.5],
                           "string_val": ['val1', 'val2', 'val3']})
        proc = FuseProcessorDataFrame(data=df, sample_desc_column='desc',
                                      columns_to_tensor={'int_val': torch.int8, 'float_val': torch.float64})

        self.assertDictEqual(proc('one'), {'int_val': torch.tensor(4, dtype=torch.int8),
                                           'float_val': torch.tensor(4.1, dtype=torch.float64), 'string_val':'val1'})
        self.assertDictEqual(proc('two'), {'int_val': torch.tensor(5, dtype=torch.int8),
                                           'float_val': torch.tensor(5.3, dtype=torch.float64), 'string_val':'val2'})
        self.assertDictEqual(proc('three'), {'int_val': torch.tensor(6, dtype=torch.int8),
                                           'float_val': torch.tensor(6.5, dtype=torch.float64), 'string_val':'val3'})


    def test_tensors_non_existing(self):
        df = pd.DataFrame({"desc": ['one', 'two', 'three'], "int_val": [4, 5, 6], "string_val": ['val1', 'val2', 'val3']})
        proc = FuseProcessorDataFrame(data=df, sample_desc_column='desc', columns_to_tensor=['invalid_column'])

        self.assertDictEqual(proc('one'), {'int_val': 4, 'string_val':'val1'})
        self.assertDictEqual(proc('two'), {'int_val': 5, 'string_val':'val2'})
        self.assertDictEqual(proc('three'), {'int_val': 6, 'string_val':'val3'})


    def tearDown(self):
        pass

if __name__ == '__main__':
    unittest.main()


"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

import logging
import traceback
from typing import Dict, Optional, Callable

import numpy as np
import sklearn.metrics
import torch
from fuse.metrics.metric_base import FuseMetricBase


def make_one_hot(input, num_classes, device='cuda'):
    """Convert class index tensor to one hot encoding tensor.
    Args:
         input: A tensor of shape [N, 1, *]
         num_classes: An int of number of class
    Returns:
        A tensor of shape [N, num_classes, *]
    """
    if len(input.shape) < 4:
        input = input.unsqueeze(0).unsqueeze(0)
    shape = np.array(input.shape)
    shape[1] = num_classes
    shape = tuple(shape)
    result = torch.zeros(shape)
    result = result.scatter_(1, input.long(), 1)
    result = result.squeeze(0).cpu().numpy()
    return result


class FuseMetricAUCPerPixel(FuseMetricBase):
    """
    Area under receiver operating characteristic curve
    Multi class version
    Will add to epoch results:
    metrics.auc.<cls> - auc per class vs rest - where <cls> will be the class name if supplied or class index
    metric.auc.avg - average of all class vs rest results
    """

    def __init__(self, pred_name: str, target_name: str, filter_func: Optional[Callable] = None, class_names: Optional[str]=None):
        """
        :param pred_name:       batch_dict key for predicted output (e.g., class probabilities after softmax)
        :param target_name:     batch_dict key for target (e.g., ground truth label)
        :param filter_func: function that filters batch_dict/ The function gets ans input batch_dict and returns filtered batch_dict
        :param class_names:     Optional - name for each class otherwise the class index will be used.
        """
        super().__init__(pred_name, target_name, filter_func)
        self._class_names = class_names

    def process(self) -> Dict[str, float]:
        """
        Calculates ROCAUC per epoch
        :return: dictionary including Area under ROC curve (floating point in range [0, 1]/-1) for each class (class vs rest),
                 -1 will be set for invalid/undefined result (an error message will be printed)
                 The dictionary will also include the average AUC
        """
        results = {}
        aucs_per_class = []
        if self._class_names is None:
            num_classes = self.epoch_preds[0].shape[0]
            class_names = ['class_' + str(i) for i in range(num_classes)]
        else:
            class_names = self._class_names
        try:
            n_classes = len(class_names)

            for cls_ind, cls in enumerate(class_names):
                predictions = []
                targets = []
                for pred, targ in zip(self.epoch_preds, self.epoch_targets):
                    pred_siz = pred.shape

                    # Convert target to one hot encoding
                    if n_classes > 1 and targ.ndim == 2:
                        targ = make_one_hot(torch.from_numpy(targ), n_classes)
                    predictions.extend(pred[cls_ind].reshape(pred_siz[1]*pred_siz[2]))
                    targets.extend(targ[cls_ind].reshape(pred_siz[1]*pred_siz[2]))

                fpr, tpr, _ = sklearn.metrics.roc_curve(targets, predictions, pos_label=1)
                auc = sklearn.metrics.auc(fpr, tpr)
                results[cls] = auc
                aucs_per_class.append(auc)

            results['macro_avg'] = np.nanmean(aucs_per_class)
        except Exception as e :
            track = traceback.format_exc()
            logging.getLogger('Fuse').error(track)
            results = {cls: -1 for cls in class_names}
            results['macro_avg'] = -1
        return results


if __name__ == '__main__':
    data = {'preds': np.array([[0.8, 0.1, 0.1],
                               [0.5, 0.3, 0.2],
                               [0.6, 0.3, 0.1],
                               [0.6, 0.1, 0.3],
                               [0.7, 0.2, 0.1],
                               [0.3, 0.2, 0.5],
                               [0.1, 0.2, 0.7],
                               [0.2, 0.6, 0.2],
                               [0.3, 0.3, 0.4],
                               [0.7, 0.2, 0.1]]),
            'targets': np.array([0, 0, 0, 0, 0, 1, 1, 1, 2, 2])}

    data = {'preds': np.array([[[0.8, 0.1, 0.1],
                               [0.2, 0.5, 0.1]],
                               [[0.8, 0.1, 0.1],
                                [0.2, 0.5, 0.1]],
                               [[0.8, 0.1, 0.1],
                                [0.2, 0.5, 0.1]],
                               [[0.8, 0.1, 0.1],
                                [0.2, 0.5, 0.1]],
                               [[0.8, 0.1, 0.1],
                                [0.2, 0.5, 0.1]],
                               [[0.8, 0.1, 0.1],
                                [0.2, 0.5, 0.1]],
                               [[0.8, 0.1, 0.1],
                                [0.2, 0.5, 0.1]],
                               [[0.8, 0.1, 0.1],
                                [0.2, 0.5, 0.1]],
                               [[0.8, 0.1, 0.1],
                                [0.2, 0.5, 0.1]],
                               [[0.8, 0.1, 0.1],
                                [0.2, 0.5, 0.1]]]),
            'targets': np.array([[[1, 0, 0],
                                  [0, 0, 0]],
                                 [[1, 0, 0],
                                  [0, 0, 0]],
                                 [[1, 0, 0],
                                  [0, 0, 0]],
                                 [[1, 0, 0],
                                  [0, 0, 0]],
                                 [[1, 0, 0],
                                  [0, 0, 0]],
                                 [[1, 0, 0],
                                  [0, 0, 0]],
                                 [[1, 0, 0],
                                  [0, 0, 0]],
                                 [[1, 0, 0],
                                  [0, 0, 0]],
                                 [[1, 0, 0],
                                  [0, 0, 0]],
                                 [[1, 0, 0],
                                  [0, 0, 0]]])}

    metric = FuseMetricAUCPerPixel(pred_name='preds', target_name='targets')
    metric.collect(data)
    res = metric.process()
    for k, v in res.items():
        print(k, '\t', v)



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

import logging
import os
from functools import partial
from multiprocessing import Manager
import torch.nn.functional as F
import torch.optim as optim
from torch.utils.data.dataloader import DataLoader

from fuse.metrics.classification.metric_auc import FuseMetricAUC
from fuse.data.augmentor.augmentor_default import FuseAugmentorDefault
from fuse.data.augmentor.augmentor_toolbox import unsqueeze_2d_to_3d, aug_op_color, aug_op_affine, squeeze_3d_to_2d, \
    rotation_in_3d
from fuse.data.dataset.dataset_generator import FuseDatasetGenerator
from fuse.data.sampler.sampler_balanced_batch import FuseSamplerBalancedBatch
from fuse.losses.loss_default import FuseLossDefault
from fuse.managers.callbacks.callback_metric_statistics import FuseMetricStatisticsCallback
from fuse.managers.callbacks.callback_tensorboard import FuseTensorboardCallback
from fuse.managers.callbacks.callback_time_statistics import FuseTimeStatisticsCallback
from fuse.managers.manager_default import FuseManagerDefault

from fuse.utils.utils_gpu import FuseUtilsGPU
from fuse.utils.utils_gpu import FuseUtilsGPU as gpu
from fuse.utils.utils_logger import fuse_logger_start
from fuse.utils.utils_param_sampler import FuseUtilsParamSamplerRandBool as RandBool
from fuse.utils.utils_param_sampler import FuseUtilsParamSamplerRandInt as RandInt
from fuse.utils.utils_param_sampler import FuseUtilsParamSamplerUniform as Uniform
from fuse.utils.utils_param_sampler import FuseUtilsParamSamplerChoice


from fuse_examples.classification.prostate_x.backbone_3d_multichannel import Fuse_model_3d_multichannel,ResNet
from fuse_examples.classification.prostate_x.patient_data_source import FuseProstateXDataSourcePatient
from fuse_examples.classification.prostate_x.processor import  FuseProstateXPatchProcessor
from fuse_examples.classification.prostate_x.tasks import FuseProstateXTask
from fuse_examples.classification.prostate_x.head_1d_classifier import FuseHead1dClassifier
from fuse_examples.classification.prostate_x.post_processor import post_processing



##########################################
# Output Paths
# ##########################################

# TODO: path to save model
root_path = ''
# TODO: path for prostateX data
root_data = ''

PATHS = {'force_reset_model_dir': False,
         # If True will reset model dir automatically - otherwise will prompt 'are you sure' message.
         'model_dir': root_path + '3D/fold0_T2_DWI_ADC_Ktrans_2/',
         'cache_dir': root_path + 'cache_fold0_T2_DWI_ADC_Ktrans/',
         'dataset_dir':root_path,

         'prostate_data_path' : root_data,
         'ktrans_path': root_data + 'ProstateXKtrains-train-fixed/',
         }

#################################
# Train Template
#################################


def train_template(paths: dict, train_common_params: dict):
    # ==============================================================================
    # Logger
    # ==============================================================================
    fuse_logger_start(output_path=paths['model_dir'], console_verbose_level=logging.INFO,
                      list_of_source_files=[])
    lgr = logging.getLogger('Fuse')
    lgr.info('Fuse Train', {'attrs': ['bold', 'underline']})

    lgr.info(f'model_dir={paths["model_dir"]}', {'color': 'magenta'})
    lgr.info(f'cache_dir={paths["cache_dir"]}', {'color': 'magenta'})

    # ==============================================================================
    # Data
    # ==============================================================================
    #### Train Data

    lgr.info(f'Train Data:', {'attrs': 'bold'})

    ## Create data source:
    DATABASE_REVISION = 29062021
    lgr.info(f'database_revision={DATABASE_REVISION}', {'color': 'magenta'})

    # create data source
    train_data_source = FuseProstateXDataSourcePatient(PATHS['dataset_dir'],'train',
                                                  db_ver=DATABASE_REVISION, db_name = train_common_params['db_name'],
                                                  fold_no = train_common_params['fold_no'], include_gt=False)

    ## Create data processors:
    image_processing_args = {
        'patch_xy': 74,
        'patch_z': 13,
    }

    ## Create data processor

    generate_processor = FuseProstateXPatchProcessor(path_to_db=PATHS['dataset_dir'],
                                                data_path=PATHS['prostate_data_path'],
                                                ktrans_data_path=PATHS['ktrans_path'],
                                                db_name=train_common_params['db_name'],
                                                fold_no=train_common_params['fold_no'],
                                                lsn_shape=(image_processing_args['patch_z'],
                                                           image_processing_args['patch_xy'],
                                                           image_processing_args['patch_xy']),
                                                                 )

    train_post_processor = partial(post_processing)

    # data augmentation (optional)
    num_channels =TRAIN_COMMON_PARAMS['backbone_model_dict']['input_channels_num']
    slice_num = image_processing_args['patch_z']

    image_channels = [list(range(0, slice_num))]
    mask_channels = [[i] for i in range(slice_num, slice_num * 2)]
    aug_pipeline = [
        [
            ('data.input',),
            rotation_in_3d,
            {'z_rot': Uniform(-5.0, 5.0), 'y_rot': Uniform(-5.0, 5.0), 'x_rot': Uniform(-5.0, 5.0)},
            {'apply': RandBool(0.5)}
        ],
        [
            ('data.input',),
            squeeze_3d_to_2d,
            {'axis_squeeze': 'z'},
            {}
        ],
        [
            ('data.input',),
            aug_op_affine,
            {'rotate': Uniform(0, 360.0),
             'translate':(RandInt(-4, 4), RandInt(-4, 4)),
             'flip': (RandBool(0.5), RandBool(0.5)),
             'scale': Uniform(0.9, 1.1),
             },
            {'apply': RandBool(0.5)}
        ],
        [
            ('data.input',),
            aug_op_affine,
            {'rotate': Uniform(-3.0, 3.0),
             'translate': (RandInt(-2, 2), RandInt(-2, 2)),
             'flip': (False, False),
             'scale': Uniform(0.9, 1.1),
             'channels': FuseUtilsParamSamplerChoice(image_channels, probabilities=None)},
            {'apply': RandBool(0.5) if train_common_params['data.aug.phase_misalignment'] else 0}
        ],

        [
            ('data.input',),
            aug_op_color,
            {'add': Uniform(-0.06, 0.06),
             'mul': Uniform(0.95, 1.05),
             'gamma': Uniform(0.9, 1.1),
             'contrast': Uniform(0.85, 1.15)},
            {'apply': RandBool(0.5)}
        ],

        [
            ('data.input',),
            unsqueeze_2d_to_3d,
            {'channels': num_channels, 'axis_squeeze': 'z'},
            {}
        ],
    ]
    augmentor = FuseAugmentorDefault(augmentation_pipeline=aug_pipeline)

    # Create dataset
    train_dataset = FuseDatasetGenerator(cache_dest=paths['cache_dir'],
                                         data_source=train_data_source,
                                         processor=generate_processor,
                                         post_processing_func=train_post_processor,
                                         augmentor=augmentor,
                                         statistic_keys=['data.ground_truth']
                                         )

    gpu_ids_for_caching = []
    lgr.info(f'- Load and cache data:')
    with Manager() as mp_manager:
        # Use gpu or cpu to generate the data - one per process - use gpus to speed it up
        gpu_list = mp_manager.list(gpu_ids_for_caching)
        train_dataset.create(num_workers=len(gpu_list), worker_init_func=FuseUtilsGPU.allocate_gpu_for_process,
                             worker_init_args=(gpu_list,))

    lgr.info(f'- Load and cache data: Done')

    #### Validation data
    lgr.info(f'Validation Data:', {'attrs': 'bold'})

    ## Create data source
    validation_data_source = FuseProstateXDataSourcePatient(PATHS['dataset_dir'],'validation',
                                                       db_ver=DATABASE_REVISION ,
                                                       db_name = train_common_params['db_name'],
                                                       fold_no=train_common_params['fold_no'], include_gt=False)

    # post processor
    validation_post_processor = partial(post_processing)

    ## Create dataset
    validation_dataset = FuseDatasetGenerator(cache_dest=paths['cache_dir'],
                                              data_source=validation_data_source,
                                              processor=generate_processor,
                                              post_processing_func=validation_post_processor,
                                              augmentor=None,
                                              statistic_keys=['data.ground_truth']
                                              )

    lgr.info(f'- Load and cache data:')
    with Manager() as mp_manager:
        # Use gpu or cpu to generate the data - one per process - use gpus to speed it up
        gpu_list = mp_manager.list(gpu_ids_for_caching)
        validation_dataset.create(num_workers=len(gpu_list), worker_init_func=FuseUtilsGPU.allocate_gpu_for_process,
                                  worker_init_args=(gpu_list,))#len(gpu_list)
    lgr.info(f'Data - task caching and filtering:', {'attrs': 'bold'})


    ## Create dataloader
    lgr.info(f'- Create sampler:')

    sampler = FuseSamplerBalancedBatch(dataset=train_dataset,
                                       balanced_class_name='data.ground_truth',
                                       num_balanced_classes=train_common_params['multi_phase_task_loss'].num_classes(),
                                       batch_size=train_common_params['data.batch_size'],
                                       balanced_class_weights=[25] * train_common_params[
                                           'multi_phase_task'].num_classes(),
                                       use_dataset_cache=True)

    lgr.info(f'- Create sampler: Done')

    # ## Create dataloader
    train_dataloader = DataLoader(dataset=train_dataset,
                                  batch_sampler=sampler,
                                  collate_fn=train_dataset.collate_fn,
                                  num_workers=train_common_params['data.train_num_workers'])
    lgr.info(f'Train Data: Done', {'attrs': 'bold'})

    validation_dataloader = DataLoader(dataset=validation_dataset,
                                       shuffle=False,
                                       drop_last=False,
                                       batch_size=train_common_params['data.batch_size'],
                                       num_workers=train_common_params['data.validation_num_workers'],
                                       collate_fn=validation_dataset.collate_fn)
    lgr.info(f'Validation Data: Done', {'attrs': 'bold'})

    # ==============================================================================
    # Model
    # ==============================================================================
    lgr.info('Model:', {'attrs': 'bold'})

    model = Fuse_model_3d_multichannel(
        conv_inputs=(('data.input', 1),),
        backbone= ResNet(),
        heads=[
        FuseHead1dClassifier(head_name='ClinSig',
                                        conv_inputs=[('model.backbone_features',  train_common_params['num_backbone_features'])],
                                        post_concat_inputs=None,
                                        dropout_rate=0.25,
                                        shared_classifier_head=None,
                                        layers_description=None,
                                        num_classes=2),

        ]
    )
    lgr.info('Model: Done', {'attrs': 'bold'})

    # ====================================================================================
    #  Loss
    # ====================================================================================
    lgr.info('Losses: CrossEntropy', {'attrs': 'bold'})

    losses = {
            'cls_loss': FuseLossDefault(pred_name='model.logits.ClinSig',
                                        target_name='data.ground_truth',
                                        callable=F.cross_entropy, weight=1.0),
        }


    # ====================================================================================
    # Metrics
    # ====================================================================================
    lgr.info('Metrics:', {'attrs': 'bold'})

    metrics = {

        'auc': FuseMetricAUC(pred_name='model.output.ClinSig', target_name='data.ground_truth',
                                class_names=train_common_params['multi_phase_task'].class_names()),
    }


    # =====================================================================================
    #  Callbacks
    # =====================================================================================
    callbacks = [
        FuseTensorboardCallback(model_dir=paths['model_dir']),  # save statistics for tensorboard
        FuseMetricStatisticsCallback(output_path=paths['model_dir'] + "/metrics.csv"),
        # save statistics for tensorboard in a csv file
        FuseTimeStatisticsCallback(num_epochs=train_common_params['manager.train_params']['num_epochs'],
                                   load_expected_part=0.1)  # time profiler
    ]

    # =====================================================================================
    #  Manager - Train
    # =====================================================================================
    lgr.info('Train:', {'attrs': 'bold'})

    # create optimizer
    optimizer = optim.Adam(model.parameters(),
                           lr=train_common_params['manager.learning_rate'],
                           weight_decay=train_common_params['manager.weight_decay'])

    # create scheduler
    scheduler = optim.lr_scheduler.ReduceLROnPlateau(optimizer, patience=3, verbose=True)

    # train from scratch
    manager = FuseManagerDefault(output_model_dir=paths['model_dir'], force_reset=paths['force_reset_model_dir'])
    # Providing the objects required for the training process.
    manager.set_objects(net=model,
                        optimizer=optimizer,
                        losses=losses,
                        metrics=metrics,
                        best_epoch_source=train_common_params['manager.best_epoch_source'],
                        lr_scheduler=scheduler,
                        callbacks=callbacks,
                        train_params=train_common_params['manager.train_params'])

    ## Continue training
    if train_common_params['manager.resume_checkpoint_filename'] is not None:
        # Loading the checkpoint including model weights, learning rate, and epoch_index.
        manager.load_checkpoint(checkpoint=train_common_params['manager.resume_checkpoint_filename'], mode='train')

    # Start training
    manager.train(train_dataloader=train_dataloader,
                  validation_dataloader=validation_dataloader)

    lgr.info('Train: Done', {'attrs': 'bold'})


######################################
# Run
######################################
if __name__ == "__main__":

    # allocate gpus
    NUM_GPUS = 1
    if NUM_GPUS == 0:
        TRAIN_COMMON_PARAMS['manager.train_params']['device'] = 'cpu'
    # uncomment if you want to use specific gpus instead of automatically looking for free ones
    force_gpus = None  # [0]
    FuseUtilsGPU.choose_and_enable_multiple_gpus(NUM_GPUS, force_gpus=force_gpus)

    ##########################################
    # Train Common Params
    ##########################################
    # ============
    # Data
    # ============
    TRAIN_COMMON_PARAMS = {}
    TRAIN_COMMON_PARAMS['db_name']='prostate_x'
    TRAIN_COMMON_PARAMS['fold_no']=0
    TRAIN_COMMON_PARAMS['data.batch_size'] = 50
    TRAIN_COMMON_PARAMS['data.train_num_workers'] = 10
    TRAIN_COMMON_PARAMS['data.validation_num_workers'] = 10
    # add misalignment to segmentation
    TRAIN_COMMON_PARAMS['data.aug.mask_misalignment'] = True
    # add misalignment to phase registration
    TRAIN_COMMON_PARAMS['data.aug.phase_misalignment'] = True

    # ===============
    # Manager - Train
    # ===============
    TRAIN_COMMON_PARAMS['manager.train_params'] = {
        'num_gpus': 1,
        'num_epochs': 150,
        'virtual_batch_size': 1,  # number of batches in one virtual batch
        'start_saving_epochs': 120,  # first epoch to start saving checkpoints from
        'gap_between_saving_epochs': 5,  # number of epochs between saved checkpoint
    }
    TRAIN_COMMON_PARAMS['manager.best_epoch_source'] = [
        {
            'source': 'metrics.auc.macro_avg',  # can be any key from losses or metrics dictionaries
            'optimization': 'max',  # can be either min/max
            'on_equal_values': 'better',
            # can be either better/worse - whether to consider best epoch when values are equal
        },

    ]
    TRAIN_COMMON_PARAMS['manager.learning_rate'] = 1e-5
    TRAIN_COMMON_PARAMS['manager.weight_decay'] = 1e-4
    TRAIN_COMMON_PARAMS['manager.dropout'] = 0.5
    TRAIN_COMMON_PARAMS['manager.momentum'] = 0.9
    TRAIN_COMMON_PARAMS['manager.resume_checkpoint_filename'] = None  # if not None, will try to load the checkpoint

    TRAIN_COMMON_PARAMS['num_backbone_features'] = 512
    TRAIN_COMMON_PARAMS['multi_phase_task_loss'] = FuseProstateXTask('ClinSig', 0)
    TRAIN_COMMON_PARAMS['multi_phase_task'] = FuseProstateXTask('ClinSig', 0)


    ##

    TRAIN_COMMON_PARAMS['mode'] = 'single_chan'  #
    # additional backbone that extract global features from a full slice which then concatenated to volume patch features
    TRAIN_COMMON_PARAMS['append_global_slice'] = False
    # global features to concat after the global max pooling
    TRAIN_COMMON_PARAMS[
        'append_features_to_gmp'] = None#[('data.tensor_clinical', 3)
    TRAIN_COMMON_PARAMS['append_features_to_image'] =  (None,0)

    TRAIN_COMMON_PARAMS['ch_inx_to_use'] = []
    if TRAIN_COMMON_PARAMS['append_global_slice']:
        TRAIN_COMMON_PARAMS['append_features_to_gmp'].append(
            ('model.backbone_global_features', TRAIN_COMMON_PARAMS['num_backbone_features']))
    # backbone parameters
    TRAIN_COMMON_PARAMS['backbone_model_dict'] = \
        {'input_channels_num': 5,
         }
    # train
    train_template(paths=PATHS, train_common_params=TRAIN_COMMON_PARAMS)



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

from typing import Dict, Tuple, Sequence, Optional

import torch
import torch.nn as nn
import torch.nn.functional as F

from fuse.utils.utils_hierarchical_dict import FuseUtilsHierarchicalDict


class FuseHead3dClassifier(nn.Module):
    """
    Model that capture slice feature including the 3D context given the local feature about a slice.
    """

    def __init__(self, head_name: str = 'head_0',
                 conv_inputs: Sequence[Tuple[str, int]] = (('model.backbone_features', 512),),
                 dropout_rate: float = 0.1,
                 num_classes: int = 3,
                 append_features: Optional[Tuple[str, int]] = None
                 ) -> None:
        """
        Create simple 3D context model
        :param head_name: string representing the head name
        :param conv_inputs: Sequence of tuples, each indication features name in batch_dict and size of features (channels)
        :param dropout_rate: dropout fraction
        :param num_classes: number of output classes
        :param append_features: Sequence of tuples, each indication features name in batch_dict and size of features (channels).
                                Those are global features that appended after the global max pooling operation

        """
        super().__init__()
        # save input params
        self.head_name = head_name
        self.conv_inputs = conv_inputs
        self.dropout_rate = dropout_rate
        self.num_classes = num_classes
        self.append_features = append_features
        self.gmp = nn.AdaptiveMaxPool3d(output_size=1)
        self.features_size = sum([features[1] for features in self.conv_inputs]) if self.conv_inputs is not None else 0

        # calc appended feature size if used
        if self.append_features is not None:
            global_features_size = sum([features[1] for features in self.append_features])
            self.features_size += global_features_size

        self.conv_classifier_3d = nn.Sequential(
            nn.Conv3d(self.features_size, 256, kernel_size=1),
            nn.ReLU(),
            nn.Dropout3d(p=self.dropout_rate),
            nn.Conv3d(256, self.num_classes, kernel_size=1),
        )

        self.do = nn.Dropout3d(p=self.dropout_rate)

    def forward(self, batch_dict: Dict) -> Dict:
        """
        Forward pass
        :param batch_dict: dictionary containing an input tensor representing spatial features with 3D context. shape: [batch_size, in_features, z, y, x]
        :return: batch dict with fields model.outputs and model.logits
        """
        if self.conv_inputs is not None:
            conv_input = torch.cat(
                [FuseUtilsHierarchicalDict.get(batch_dict, conv_input[0]) for conv_input in self.conv_inputs], dim=1)
            global_features = self.gmp(conv_input)
            # save global max pooling features in case needed (mostly to analyze)
            FuseUtilsHierarchicalDict.set(batch_dict, 'model.' + self.head_name +'.gmp_features', global_features.squeeze(dim=4).squeeze(dim=3).squeeze(dim=2))
            # backward compatibility
            if hasattr(self, 'do'):
                global_features = self.do(global_features)
        # append global features if are used
        if self.append_features is not None:
            features = torch.cat(
                [FuseUtilsHierarchicalDict.get(batch_dict, features[0]).reshape(-1, features[1]) for features in self.append_features], dim=1)
            features = features.reshape(features.shape + (1,1,1))
            if self.conv_inputs is not None:
                global_features = torch.cat((global_features, features), dim=1)
            else:
                global_features = features

        logits = self.conv_classifier_3d(global_features)
        logits = logits.squeeze(dim=4)
        logits = logits.squeeze(dim=3)
        logits = logits.squeeze(dim=2)  # squeeze will change the shape to  [batch_size, channels']

        cls_preds = F.softmax(logits, dim=1)
        FuseUtilsHierarchicalDict.set(batch_dict, 'model.logits.' + self.head_name, logits)
        FuseUtilsHierarchicalDict.set(batch_dict, 'model.output.' + self.head_name, cls_preds)

        return batch_dict



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

from typing import Sequence, Dict, Tuple, Callable, Optional

import torch

from fuse.models.backbones.backbone_inception_resnet_v2 import FuseBackboneInceptionResnetV2
from fuse.models.heads.head_global_pooling_classifier import FuseHeadGlobalPoolingClassifier
from fuse.utils.utils_hierarchical_dict import FuseUtilsHierarchicalDict


class FuseModelMultistream(torch.nn.Module):
    """
    Multi-stream Fuse model - convolutional neural network with multiple processing streams and multiple heads
    """

    def __init__(self,
                 conv_inputs: Tuple[str, int] = ('data.input.input_0.tensor', 1),
                 backbone_streams: Sequence[torch.nn.Module] = (FuseBackboneInceptionResnetV2(logical_units_num=12),
                                                                FuseBackboneInceptionResnetV2(logical_units_num=12)),
                 heads: Sequence[torch.nn.Module] = (FuseHeadGlobalPoolingClassifier(),),
                 split_logic: Optional[Callable] = None,
                 join_logic: Optional[Callable] = None,
                 ) -> None:
        """
        Multi-stream Fuse model - convolutional neural network with multiple processing streams and multiple heads
        :param conv_inputs:             batch_dict name for model input and its number of input channels
        :param backbone_streams:        List of PyTorch backbone modules - one per stream (can share weights or not)
        :param heads:                   Sequence of head modules
        :param split_logic:             Optional callable, splits input into streams. If None, sends each input channel to consecutive stream.
                                            Signature: stream_outputs = split_logic(batch_dict, backbone_streams)
        :param join_logic:              Optional callable, joins stream outputs into single feature map. If None, concatenates on channel axis.
                                            Signature: feature_map = join_logic(batch_dict, stream_outputs)
        """
        super().__init__()
        self.conv_inputs = conv_inputs
        self.split_logic = split_logic
        self.join_logic = join_logic

        # Register modules
        self.backbone_streams = torch.nn.ModuleList(backbone_streams)
        self.add_module('backbones', self.backbone_streams)
        self.heads = torch.nn.ModuleList(heads)
        self.add_module('heads', self.heads)

    def forward(self,
                batch_dict: Dict) -> Dict:

        # Forward pass through multiple streams
        # -------------------------------------
        if self.split_logic is None:
            # If no split logic is provided, send each channel to different stream
            conv_input = FuseUtilsHierarchicalDict.get(batch_dict, self.conv_inputs[0])  # shape = [batch_size, num_channels, height, width]
            stream_outputs = []
            for ch_idx in range(conv_input.shape[1]):
                single_channel_batch = conv_input[:, ch_idx, :, :].unsqueeze(dim=1)  # shape = [batch_size, 1, height, width]
                stream_output = self.backbone_streams[ch_idx](single_channel_batch)
                stream_outputs.append(stream_output)
        elif callable(self.split_logic):
            stream_outputs = self.split_logic(batch_dict, self.backbone_streams)
        else:
            raise Exception('Error in FuseModelMultistream - bad split logic provided')

        # Combining feature maps from multiple streams
        # --------------------------------------------
        if self.join_logic is None:
            # If no join logic is provided, concatenate feature maps in channel axis
            backbone_features = torch.cat(stream_outputs, dim=1)
        elif callable(self.join_logic):
            backbone_features = self.join_logic(batch_dict, stream_outputs)
        else:
            raise Exception('Error in FuseModelMultistream - bad join logic provided')

        FuseUtilsHierarchicalDict.set(batch_dict, 'model.backbone_features', backbone_features)
        for head in self.heads:
            batch_dict = head.forward(batch_dict)

        return batch_dict['model']


if __name__ == '__main__':
    from fuse.models.heads.head_dense_segmentation import FuseHeadDenseSegmentation

    backbone_0 = FuseBackboneInceptionResnetV2(logical_units_num=8)
    backbone_1 = FuseBackboneInceptionResnetV2(logical_units_num=8)

    non_shared_model = FuseModelMultistream(
        conv_inputs=('data.input.input_0.tensor', 2),
        backbone_streams=[backbone_0, backbone_1],
        heads=[
            FuseHeadGlobalPoolingClassifier(head_name='head_0',
                                            conv_inputs=[('model.backbone_features', 640)],
                                            post_concat_inputs=None,
                                            num_classes=2),

            FuseHeadDenseSegmentation(head_name='head_1',
                                      conv_inputs=[('model.backbone_features', 640)],
                                      num_classes=2)
        ]
    )

    shared_model = FuseModelMultistream(
        conv_inputs=('data.input.input_0.tensor', 2),
        backbone_streams=[backbone_0, backbone_0],
        heads=[
            FuseHeadGlobalPoolingClassifier(head_name='head_0',
                                            conv_inputs=[('model.backbone_features', 640)],
                                            post_concat_inputs=None,
                                            num_classes=2),

            FuseHeadDenseSegmentation(head_name='head_1',
                                      conv_inputs=[('model.backbone_features', 640)],
                                      num_classes=2)
        ]
    )

    dummy_data = {'data':
                      {'input':
                           {'input_0': {'tensor': torch.zeros([20, 2, 200, 100]), 'metadata': None}},
                       'gt':
                           {'gt_0': torch.zeros([0])}
                       }
                  }

    for model_name, model in [('Shared-model', shared_model), ('Non-shared-model', non_shared_model)]:
        res = {}
        res['model'] = shared_model.forward(dummy_data)
        print(model_name + ', forward pass shape - head_0: ', end='')
        print(str(res['model']['logits']['head_0'].shape))
        print(model_name + ', forward pass shape - head_1: ', end='')
        print(str(res['model']['logits']['head_1'].shape))
        total_params = sum(p.numel() for p in model.parameters())
        backbone_params = sum(p.numel() for p in model._modules['backbones'].parameters())
        print(model_name + ', backbone params = %d' % backbone_params)
        print(model_name + ', heads params = %d' % (total_params - backbone_params))
        print(model_name + ', total params = %d' % total_params)
        print()



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

from typing import Dict

from fuse.managers.manager_state import FuseManagerState


class FuseCallback(object):
    """
        Abstract base class used to build new callbacks.
        Callbacks are called at various stages during training and infer.

    """

    def __init__(self):
        pass

    def on_step_begin(self, step: int) -> None:
        """
        In general, a step is the part of the train that: handle_epoch for both train and validation data, update the scheduler and save checkpoint.
        The step number is the epoch number.

        Called on step begin.
        :param step: step number
        """
        pass

    def on_step_end(self, step: int, train_results: Dict = None, validation_results: Dict = None, learning_rate: float = None) -> None:
        """
        In general, a step is the part of the train that:
        handle_epoch for both train and validation data, update the scheduler and save checkpoint.
        The step number is the epoch number.

        Called on step end.
        :param step: step number
        :param train_results: contains train losses and metrics.
                    e.g.: {
                        'losses': {'cls_loss': 0.53242,
                                   'total_loss': 0.53242},
                        'metrics' : {'auc' : 0.8676}
                        }
        :param validation_results: contains validation losses and metrics.
                    e.g.: {
                        'losses': {'cls_loss': 0.53242,
                                   'total_loss': 0.53242},
                        'metrics' : {'auc' : 0.8676}
                        }
        :param learning_rate: the learning rate at the end of the step.
        """
        pass

    def on_epoch_begin(self, mode: str, epoch: int) -> None:
        """
        Called at the beginning of epoch.

        :param mode: either 'train', 'validation' or 'infer'
        :param epoch: epoch number
        """
        pass

    def on_epoch_end(self, mode: str, epoch: int, epoch_results: Dict = None) -> None:
        """
        Called at the end of epoch handling.

        :param mode: either 'train', 'validation' or 'infer'
        :param epoch: epoch number
        :param epoch_results: hierarchical dictionary where keys are losses and metrics computed on data.
            E.g.: {
                    'losses': {'loss1': mean_loss1,
                               'loss2': mean_loss2,
                                'total_loss': mean (loss1 + loss2)}
                    'metrics': {'metric1': epoch_metric1,
                                'metric2': epoch_metric2}
               }
            in 'infer' mode it is an empty dict {}
        """
        pass

    def on_virtual_batch_begin(self, mode: str, virtual_batch: int) -> None:
        """
        Called at the beginning of virtual_batch.

        :param mode: either 'train', 'validation' or 'infer'
        :param virtual_batch: virtual batch number
        """
        pass

    def on_virtual_batch_end(self, mode: str, virtual_batch: int, virtual_batch_results: Dict = None) -> None:
        """

        :param mode: either 'train', 'validation' or 'infer'
        :param virtual_batch: virtual batch number
        :param virtual_batch_results: hierarchical dictionary with the computed losses
                    'losses': {'loss1': mean_loss1,
                               'loss2': mean_loss2,
                               'total_loss': mean (loss1 + loss2)}
        """
        pass

    def on_batch_begin(self, mode: str, batch: int) -> None:
        pass

    def on_data_fetch_end(self, mode: str, batch: int, batch_dict: Dict = None) -> None:
        """
        Called after iter.next() is called.

        :param mode: either 'train', 'validation' or 'infer'
        :param batch: batch number
        :param batch_dict: dict with the retrieved data:
            it's key is 'data' and it contains gt, input and descriptor keys
        """
        pass

    def on_batch_end(self, mode: str, batch: int, batch_dict: Dict = None) -> None:
        """
        Called at the end of handling batch.

        :param mode: either 'train', 'validation' or 'infer'
        :param batch: batch number
        :param batch_dict:
            hierarchical dictionary contains the keys:
            data - with gt, input and descriptor keys.
            losses - key for each computed loss + total_loss.
        """
        pass

    def on_train_begin(self, state: FuseManagerState) -> None:
        """
        Called at the beginning of the train procedure, after initialization of all variables and model.

        :param state: manager state object.
            Contains the state of the manager. For details, see FuseManagerState.
        """
        pass

    def on_train_end(self) -> None:
        pass



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

from typing import Union, Sequence, Dict, Tuple

from torch.utils.data import Dataset

from fuse.data.data_source.data_source_from_list import FuseDataSourceFromList
from fuse.data.dataset.dataset_default import FuseDatasetDefault
from fuse.data.processor.processor_base import FuseProcessorBase


# Dataset processor
class DatasetProcessor(FuseProcessorBase):
    """
    Processor that extract data from pytorch dataset and convert each sample to dictionary
    """

    def __init__(self, dataset: Dataset, mapping: Sequence[str]):
        """
        :param dataset: the pytorch dataset to convert
        :param mapping: dictionary key for each element returned by the pytorch dataset
        """
        # store input arguments
        self.mapping = mapping
        self.dataset = dataset

    def __call__(self, desc: Tuple[str, int], *args, **kwargs):
        index = desc[1]
        sample = self.dataset[index]
        sample = {self.mapping[i]: val for i, val in enumerate(sample)}

        return sample


class FuseDatasetWrapper(FuseDatasetDefault):
    """
    Fuse Dataset Wrapper
    wraps pytorch dataset.
    Each sample will be converted to dictionary according to mapping.
    And this dataset inherits all FuseDatasetDefault features
    """

    #### CONSTRUCTOR
    def __init__(self, name: str, dataset: Dataset, mapping: Union[Sequence, Dict[str, str]], **kwargs):
        """
        :param name: name of the data extracted from dataset, typically: 'train', 'validation;, 'test'
        :param dataset: the dataset to extract the data from
        :param mapping: including name for each returned object from dataset
        :param kwargs: optinal, additional argumentes to provide to FuseDatasetDefault
        """
        data_source = FuseDataSourceFromList([(name, i) for i in range(len(dataset))])
        processor = DatasetProcessor(dataset, mapping)
        super().__init__(data_source=data_source, input_processors=None, gt_processors=None,processors=processor,  **kwargs)



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

"""
Cache to file per sample
"""
import gzip
import logging
import os
import pickle
import traceback
from multiprocessing import Manager
from typing import Hashable, Any, List
import torch
torch.multiprocessing.set_sharing_strategy('file_system')

from fuse.data.cache.cache_base import FuseCacheBase
from fuse.utils.utils_atomic_file import FuseUtilsAtomicFileWriter
from fuse.utils.utils_file import FuseUtilsFile


class FuseCacheFiles(FuseCacheBase):
    def __init__(self, cache_file_dir: str, reset_cache: bool, single_file: bool=False):
        """
        :param cache_file_dir: path to cache dir
        :param reset_cache: reset previous cache if exist or continue
        """
        super().__init__()

        self._cache_file_dir = cache_file_dir

        # create dir if not already exist
        FuseUtilsFile.create_dir(cache_file_dir)

        # pointer to cache index
        self._cache_file_name = os.path.join(self._cache_file_dir, 'cache_index.pkl')
        self._cache_prop_file_name = os.path.join(self._cache_file_dir, 'cache_properties.pkl')

        # reset or load from disk
        if reset_cache or not os.path.exists(self._cache_file_name):
            self.reset()
            self.single_file = single_file
            # save initial properties
            with FuseUtilsAtomicFileWriter(filename=self._cache_prop_file_name) as cache_prop_file:
                pickle.dump({'single_file': self.single_file}, cache_prop_file)
        else:
            # get last modified time of the index
            self._cache_index_mtime = os.path.getmtime(self._cache_file_name)

            # load current cache
            try:
                with open(self._cache_file_name, 'rb') as cache_index_file:
                    self._cache_index = pickle.load(cache_index_file)
            except:
                # backward compatibility - used to be saved in gz format
                with gzip.open(self._cache_file_name, 'rb') as cache_index_file:
                    self._cache_index = pickle.load(cache_index_file)
            self._cache_list = list(self._cache_index.keys())

            # load mode for backward compatibility
            try:
                with open(self._cache_prop_file_name, 'rb') as cache_prop_file:
                    cache_prop = pickle.load(cache_prop_file)
                    self.single_file = cache_prop['single_file']
            except:
                self.single_file = False

    def __contains__(self, key: Hashable) -> bool:
        """
        See base class
        """
        return key in self._cache_index

    def __getitem__(self, key: Hashable) -> Any:
        """
        See base class
        """
        if self.single_file:
            return self._cache_index.get(key, None)

        value_file_name = self._cache_index.get(key, None)
        if value_file_name is None:
            return None
        value_file_name = os.path.join(self._cache_file_dir, value_file_name)

        # make sure file not exist
        if os.path.exists(value_file_name):
            # store the file
            with gzip.open(value_file_name, 'rb') as value_file:
                value = pickle.load(value_file)
        else:
            raise Exception(f'cache file {value_file_name} not found')

        return value

    def __delitem__(self, key: Hashable) -> None:
        """
        Not supported
        """
        raise NotImplementedError

    def __setitem__(self, key: Hashable, value: Any) -> None:
        """
        See base class
        """
        if not self._cache_enable:
            raise Exception('First start caching using function start_caching()')

        self._cache_list.append(key)

        # if value is none, just update cache index
        if value is None:
            self._cache_index[key] = None
            return
        if self.single_file:
            self._cache_index[key] = value
        else:
            index = self._cache_list.index(key)
            value_file_name = str(index).zfill(10) + '.pkl.gz'
            value_abs_file_name = os.path.join(self._cache_file_dir, value_file_name)
            self._cache_index[key] = value_file_name

            # make sure file not exist
            if os.path.exists(value_abs_file_name):
                logging.getLogger('Fuse').warning(f'cache file {value_abs_file_name} unexpectedly exist, overriding it.')

            # store the file
            with FuseUtilsAtomicFileWriter(value_abs_file_name) as value_file:
                pickle.dump(value, value_file)

            # store the cache index - just for a case of crashing
            try:
                with FuseUtilsAtomicFileWriter(filename=self._cache_file_name) as cache_index_file:
                    pickle.dump(dict(self._cache_index), cache_index_file)
            except:
                # do not trow error- just print warning
                lgr = logging.getLogger('Fuse')
                track = traceback.format_exc()
                lgr.warning(track)

    def save(self) -> None:
        """
        Save cache index file
        """
        # disable caching
        self._cache_enable = False

        with FuseUtilsAtomicFileWriter(filename=self._cache_file_name) as cache_index_file:
            pickle.dump(dict(self._cache_index), cache_index_file)

        # move back to simple data structures
        self._cache_index = dict(self._cache_index)
        self._cache_list = list(self._cache_list)

    def exist(self) -> bool:
        """
        See base class
        """
        return bool(self._cache_index)

    def reset(self) -> None:
        """
        See base class
        """
        # make sure the dir content is empty
        FuseUtilsFile.remove_dir_content(self._cache_file_dir)

        # create empty data structures
        self._cache_enable = False
        self._cache_index = {}
        self._cache_list = []
        self._cache_index_mtime = -1

    def get_all_keys(self, include_none: bool = False) -> List[Hashable]:
        """
        See base class
        """
        if include_none:
            return list(self._cache_index.keys())
        else:
            return [key for key, value in self._cache_index.items() if value is not None]

    def start_caching(self, manager: Manager):
        """
        See base class
        """
        self._cache_enable = True
        # if manager is  None assume that the it's not multiprocessing caching
        if manager is not None:
            # create dictionary and adds it one by one to workaround multiprocessing limitation
            cache_index = manager.dict()
            for k, v in self._cache_index.items():
                cache_index[k] = v
            self._cache_index = cache_index
            self._cache_list = manager.list(self._cache_list)



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

from typing import List, Tuple

from fuse.data.data_source.data_source_base import FuseDataSourceBase
from fuse_examples.classification.prostate_x.data_utils import FuseProstateXUtilsData

class FuseProstateXDataSourcePatient(FuseDataSourceBase):
    def __init__(self,
                 db_path: str,
                 set_type: str,
                 db_name: str,
                 db_ver:int = 11,
                 fold_no: int=0,
                 include_gt: bool = True,
                 include_pred: bool = True,

                 ):
        """
        Fuse DataSource for ProstateX data.
        Generate sample decription per patient

        :param  set_type: 'train' 'validation' 'test'
        :param db_ver: database version
        :type include_gt: create two descriptors per patient - with 'gt' key and without
        :return list of sample descriptors

        """
        self.db_path = db_path
        self.set_type = set_type
        self.db_ver = db_ver
        self.include_gt = include_gt
        self.include_pred = include_pred
        self.db_name = db_name
        self.fold_no = fold_no
        self.desc_list = self.generate_patient_list()




    def get_samples_description(self):
        return list(self.desc_list)

    def summary(self) -> str:
        """
        See base class
        """
        summary_str = ''
        summary_str += f'Class = {type(self)}\n'
        summary_str += f'Input source = {self.set_type}\n'
        summary_str += f'Number of Patients = {len(self.desc_list)}\n'
        return summary_str


    def generate_patient_list(self) -> List[Tuple]:
        '''
        Go Over all patients and create a tuple list of (db_ver, set_type, patient_id [,'gt'])
        :return: list of patient descriptors
        '''
        data = FuseProstateXUtilsData.get_dataset(self.db_path,self.set_type, self.db_ver,self.db_name,self.fold_no)
        if self.db_name=='prostate_x':
            data_lesions = FuseProstateXUtilsData.get_lesions_prostate_x(data)

        patients = list(data_lesions['Patient ID'].unique())
        desc_list = []
        if self.include_pred:
            desc_list += [(self.db_ver, self.set_type, patient_id, 'pred') for patient_id in patients]
        if self.include_gt:
            desc_list += [(self.db_ver, self.set_type, patient_id, 'gt') for patient_id in patients]
        return desc_list

if __name__ == "__main__":
    path_to_db = '/gpfs/haifa/projects/m/msieve_dev3/usr/Tal/my_research/virtual_biopsy/prostate/experiments/V1/'
    train_data_source = FuseProstateXDataSourcePatient(path_to_db,'train',db_name='tcia', db_ver='18042021',fold_no=0, include_gt=False)



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

from copy import deepcopy
from enum import Enum

import numpy as np
import scipy.stats
from typing import Dict, Tuple, Optional

from fuse.metrics.metric_base import FuseMetricBase


class CIMethod(Enum):
    """
    Method to compute confidence interval from a list of bootstrap results
    """
    NORMAL = 0
    PIVOTAL = 1
    PERCENTILE = 2

    def compute_confidence_interval(self, org_statistic: np.ndarray, bootstrap_statistics: np.ndarray, confidence: float) -> Tuple[float, float]:
        lower_confidence_ratio = (1 - confidence / 100.0) / 2.0
        if self == CIMethod.NORMAL:
            sigma = np.std(bootstrap_statistics)
            c_alpha = scipy.stats.norm.isf(lower_confidence_ratio)
            return org_statistic - c_alpha * sigma, org_statistic + c_alpha * sigma
        statistic_observations = np.sort(bootstrap_statistics)

        n = len(statistic_observations)
        index_low = int(np.floor(lower_confidence_ratio * n))
        index_high = int(np.floor((1 - lower_confidence_ratio) * n))

        bootstrap_statistic_low = statistic_observations[index_low]
        bootstrap_statistic_high = statistic_observations[index_high]
        if self == CIMethod.PERCENTILE:
            return bootstrap_statistic_low, bootstrap_statistic_high

        assert self == CIMethod.PIVOTAL
        return 2 * org_statistic - bootstrap_statistic_high, 2 * org_statistic - bootstrap_statistic_low


class FuseMetricConfidenceInterval(FuseMetricBase):
    """
    Wrapper Metric to compute the confidence interval of another metric
    """

    def __init__(self,
                 metric: FuseMetricBase,
                 num_of_bootstraps: int = 10000,
                 rnd_seed: int = 1234,
                 conf_interval: float = 95,
                 stratum_name: Optional[str] = None,
                 ci_method: CIMethod = CIMethod.PERCENTILE) -> None:
        """
        :param metric: metric to compute the confidence interval for
        :param num_of_bootstraps: number of bootstrapping
        :param rng_seed: seed for random number generator.
        :param conf_interval: Confidence interval. Default is 95.
        :param stratum_name: if sampling should be done by strata, specify the key in batch_dict to collect data from
        :param ci_method: specifies the method for computing the confidence intervals from bootstrap samples

        """
        self._metric = metric
        if stratum_name is not None:
            self._metric.add_key_to_collect(name='stratum', key=stratum_name)
        self._num_of_bootstraps = num_of_bootstraps
        self._rnd_seed = rnd_seed
        self._conf_interval = conf_interval
        self._ci_method = ci_method

    def reset(self) -> None:
        """
        Resets collected data for metrics
        :return: None
        """
        self._metric.reset()
        pass

    def collect(self,
                batch_dict: Dict) -> None:
        """
        Calls the metric collect, to gather the data form the batch
        :param batch_dict:
        """
        self._metric.collect(batch_dict)
        pass

    def process(self) -> Dict[str, float]:
        """
        Calculate Confidence Interval for the metric

        :return: dictionary including Area under ROC curve (floating point in range [0, 1]/-1) for each class (class vs rest),
                 -1 will be set for invalid/undefined result (an error message will be printed)
                 The dictionary will also include the average AUC
        """

        def _compute_stats(orig, samples):
            confidence_lower, confidence_upper = self._ci_method.compute_confidence_interval(orig, samples, self._conf_interval)
            return {'org': orig, 'mean': np.mean(samples), 'std': np.std(samples),
                    'conf_interval': self._conf_interval, 'conf_lower': confidence_lower, 'conf_upper': confidence_upper}

        rnd = np.random.RandomState(self._rnd_seed)
        original_sample_results = self._metric.process()
        boot_results = []
        ci_results = {}

        sampled_metric: FuseMetricBase = deepcopy(self._metric)
        sampled_data = sampled_metric.collected_data
        orig_data = {}
        # make sure data is saved in arrays, also keep the original data in array
        for key, data in sampled_data.items():
            orig_data[key] = np.array(data)
            sampled_data[key] = np.empty(orig_data[key].shape)

        stratum_id = orig_data['stratum'] if 'stratum' in orig_data else np.ones(len(orig_data['target']))
        unique_strata = np.unique(stratum_id)

        for bt in range(self._num_of_bootstraps):
            for stratum in unique_strata:
                stratum_filter = stratum_id == stratum
                n_stratum = sum(stratum_filter)
                sample_ix = rnd.randint(0, n_stratum, size=n_stratum)
                for key, data in sampled_data.items():
                    sampled_data[key][stratum_filter] = orig_data[key][stratum_filter][sample_ix]
            boot_results.append(sampled_metric.process())
        # results can be either a list of floats or a list of dictionaries
        if isinstance(original_sample_results, dict):
            for key, orig_val in original_sample_results.items():
                sampled_vals = [sample[key] for sample in boot_results]
                ci_results[key] = _compute_stats(orig_val, sampled_vals)
        elif isinstance(original_sample_results, float):
            ci_results[key] = _compute_stats(original_sample_results, boot_results)

        return ci_results


if __name__ == '__main__':
    from fuse.metrics.classification.metric_auc import FuseMetricAUC
    from fuse.metrics.classification.metric_accuracy import FuseMetricAccuracy

    data = {'preds': np.array([[0.8, 0.1, 0.1],
                               [0.5, 0.3, 0.2],
                               [0.6, 0.3, 0.1],
                               [0.6, 0.1, 0.3],
                               [0.7, 0.2, 0.1],
                               [0.3, 0.2, 0.5],
                               [0.1, 0.2, 0.7],
                               [0.2, 0.6, 0.2],
                               [0.3, 0.3, 0.4],
                               [0.7, 0.2, 0.1],
                               [0.3, 0.2, 0.5],
                               [0.1, 0.2, 0.7],
                               [0.2, 0.6, 0.2],
                               [0.3, 0.3, 0.4],
                               [0.7, 0.2, 0.1]]),
            'targets': np.array([0, 0, 0, 0, 0, 1, 1, 2, 0, 0, 1, 1, 2, 2, 2]),
            'stratum': np.array([1, 0, 1, 0, 0, 1, 0, 1, 0, 0, 1, 0, 1, 0, 1])}

    auc_metric = FuseMetricAUC(pred_name='preds', target_name='targets')
    metric = FuseMetricConfidenceInterval(auc_metric, num_of_bootstraps=10, rnd_seed=198654, stratum_name='stratum')
    metric.collect(data)
    res = metric.process()
    print("AUC with stratum")
    for k, v in res.items():
        print(k, '\t', v)

    auc_metric = FuseMetricAUC(pred_name='preds', target_name='targets')
    metric = FuseMetricConfidenceInterval(auc_metric, num_of_bootstraps=10, rnd_seed=198654)
    metric.collect(data)
    res = metric.process()
    print("AUC no stratum")
    for k, v in res.items():
        print(k, '\t', v)

    data = {'preds': np.array([[0.8, 0.1, 0.1],
                               [0.5, 0.3, 0.2],
                               [0.6, 0.3, 0.1],
                               [0.6, 0.1, 0.3],
                               [0.7, 0.2, 0.1],
                               [0.3, 0.2, 0.5],
                               [0.1, 0.2, 0.7],
                               [0.2, 0.6, 0.2],
                               [0.3, 0.3, 0.4],
                               [0.7, 0.2, 0.1]]),
            'targets': np.array([0, 0, 0, 0, 0, 1, 1, 2, 2, 2]),
            'weights': np.array([0.03, 0.9, 0.05, 0.52, 0.23, 0.72, 0.13, 0.113, 0.84, 0.09])}

    accuracy_metric = FuseMetricAccuracy(pred_name='preds', target_name='targets', use_sample_weights=True,
                                         sample_weight_name='weights')
    metric = FuseMetricConfidenceInterval(accuracy_metric, 10, 5)
    metric.collect(data)
    res = metric.process()
    print("Accuracy")
    for k, v in res.items():
        print(k, '\t', v)

    data = {'preds': np.array([[0.8, 0.1, 0.1],
                               [0.5, 0.3, 0.2],
                               [0.6, 0.3, 0.1],
                               [0.6, 0.1, 0.3],
                               [0.7, 0.2, 0.1],
                               [0.3, 0.2, 0.5],
                               [0.1, 0.2, 0.7],
                               [0.2, 0.6, 0.2],
                               [0.3, 0.3, 0.4],
                               [0.7, 0.2, 0.1]]),
            'targets': np.array([0, 0, 0, 0, 0, 1, 1, 2, 2, 2]),
            'weights': np.array([0.03, 0.9, 0.05, 0.52, 0.23, 0.72, 0.13, 0.113, 0.84, 0.09]),
            'stratum': np.array([1, 0, 1, 0, 0, 1, 0, 1, 0, 1])}

    accuracy_metric = FuseMetricAccuracy(pred_name='preds', target_name='targets')
    metric = FuseMetricConfidenceInterval(accuracy_metric, 10, 5, stratum_name='stratum')
    metric.collect(data)
    res = metric.process()
    print("Accuracy with stratum, no weights")
    for k, v in res.items():
        print(k, '\t', v)


"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

import logging
import traceback
from typing import Callable, Optional, Dict

import numpy as np

from fuse.metrics.metric_base import FuseMetricBase


class FuseMetricDefault(FuseMetricBase):
    """
    Applying a given metric function
    """

    def __init__(self,
                 metric_func: Callable,
                 pred_name: str,
                 target_name: str,
                 class_names: Optional[str] = None,
                 filter_func: Optional[Callable] = None,
                 use_sample_weights: Optional[bool] = False,
                 sample_weight_name: Optional[str] = None) -> None:
        """
        :param metric_func:         metric function. Will be applied for each class in a one-vs-rest manner.
                                    Gets as an input y_true (a boolean np.ndarray), y_score and sample_weight.
        :param pred_name:           batch_dict key for predicted output (e.g., class probabilities after softmax)
        :param target_name:         batch_dict key for target (e.g., ground truth label)
        :param class_names:         Optional - name for each class otherwise the class index will be used.
        :param filter_func:         Optional - function that filters batch_dict/ The function gets ans input batch_dict and returns filtered batch_dict
        :param use_sample_weights:  Optional - when True (default is False), metrics are computed with sample weigths.
                                    For this, use argument 'sample_weight_name' to define the batch_dict key for the sample weight
        :param sample_weight_name:  Optional, when use_sample_weights is True, user must specify the key for collecting sample weights
        """
        super().__init__(pred_name, target_name, filter_func=filter_func,
                         use_sample_weights=use_sample_weights, sample_weight_name=sample_weight_name)
        self._class_names = class_names
        self.metric_func = metric_func

    def process(self) -> Dict[str, float]:
        """
        Computes an evaluation metric per epoch and class - using metric_func
        :return: dictionary including the computed metric for each class (one vs rest),
                 -1 will be set for invalid/undefined result (an error message will be printed)
                 The dictionary will also include the (unweighted) average metric across all classes
        """
        results = {}
        aucs_per_class = []
        if self._class_names is None:
            num_classes = self.collected_data['pred'][0].shape[0]
            class_names = ['class_' + str(i) for i in range(num_classes)]
        else:
            class_names = self._class_names
        sample_weight = self.collected_data['sample_weight'] if self.use_sample_weights else None
        try:
            for cls_ind, cls in enumerate(class_names):
                predicted_ind = [pred[cls_ind] for pred in self.collected_data['pred']]

                computed_metric = self.metric_func(np.asarray(self.collected_data['target']) == cls_ind,
                                                   np.asarray(predicted_ind), sample_weight=sample_weight)
                results[cls] = computed_metric
                aucs_per_class.append(computed_metric)

            results['macro_avg'] = np.nanmean(aucs_per_class)
        except:
            track = traceback.format_exc()
            logging.getLogger('Fuse').error(track)
            results = {cls: -1 for cls in class_names}
            results['macro_avg'] = -1
        return results


if __name__ == '__main__':
    from sklearn import metrics

    data = {'preds': np.array([[0.8, 0.1, 0.1],
                               [0.5, 0.3, 0.2],
                               [0.6, 0.3, 0.1],
                               [0.6, 0.1, 0.3],
                               [0.7, 0.2, 0.1],
                               [0.3, 0.2, 0.5],
                               [0.1, 0.2, 0.7],
                               [0.2, 0.6, 0.2],
                               [0.3, 0.3, 0.4],
                               [0.7, 0.2, 0.1]]),
            'targets': np.array([0, 0, 0, 0, 0, 1, 1, 1, 2, 2]),
            'weights': np.array([0.03, 0.12, 0.05, 0.52, 0.23, 0.72, 0.13, 0.113, 0.84, 0.09])}


    def f1_score(y_true, y_score, sample_weight):
        return metrics.f1_score(y_true=y_true, y_pred=y_score > 0.5, sample_weight=sample_weight)


    metric_dict = {
        "average_precision": FuseMetricDefault(metric_func=metrics.average_precision_score, pred_name='preds',
                                               target_name='targets',
                                               use_sample_weights=True,
                                               sample_weight_name='weights'),

        "f1 score": FuseMetricDefault(metric_func=f1_score, pred_name='preds',
                                      target_name='targets',
                                      use_sample_weights=True,
                                      sample_weight_name='weights'),

    }

    for metric_name, metric in metric_dict.items():
        print("**** metric=", metric_name)
        metric.collect(data)

        print("use_sample_weights = True")
        res = metric.process()
        for k, v in res.items():
            print(k, '\t', v)

        print("use_sample_weights = False")
        metric.use_sample_weights = False
        res = metric.process()
        for k, v in res.items():
            print(k, '\t', v)



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

from typing import Optional, Union, Dict, Sequence

import numpy as np

from fuse.metrics.metric_base import FuseMetricBase
from Fuse.utils.utils_image_processing import FuseUtilsImageProcessing


class FuseMetricScoreMap(FuseMetricBase):
    """
    Segmentation metrics.
    """

    def __init__(self,
                 pred_name: str,
                 target_name: str,
                 hard_threshold: bool = False,
                 threshold: float = 0.0,
                 class_names: Optional[Sequence[str]] = None):
        """
        :param pred_name:       batch_dict key for predicted output (e.g., class probabilities after softmax).
                                Expected Tensor shape = [batch, num_classes, height, width]
        :param target_name:     batch_dict key for target (e.g., ground truth label). Expected Tensor shape = [batch, height, width]
        :param hard_threshold:  Boolean for using predictions as probabilities for dice calculations of turning predicitons to binary output.
        :param threshold:       Threshold for creating a boolean image from probabilities.

        :param class_names:     Optional - class names
        """
        super().__init__(pred_name, target_name)
        self._class_names = class_names
        self.hard_threshold = hard_threshold
        self.threshold = threshold

    def process(self) -> Union[float, Dict[str, float]]:
        num_classes = self.collected_data['pred'][0].shape[0]

        raw_metrics_per_class = []  # tp/fp/fn accumulators per class
        for cls_idx in range(num_classes + 1):  # Adding one extra class for average
            raw_metrics_per_class.append({'tp': 0, 'fp': 0, 'fn': 0})

        for sample_idx in range(len(self.collected_data['pred'])):
            targets = self.collected_data['target'][sample_idx]  # shape = [height, width]
            preds = self.collected_data['pred'][sample_idx]  # shape = [num_classes, height, width]

            targets_height, targets_width = targets.shape
            num_classes, preds_height, preds_width = preds.shape

            # Resize targets (if needed) to match score map
            # ==============================================
            if (targets_height > preds_height) or (targets_width > preds_width):
                targets = FuseUtilsImageProcessing.preserve_range_resize(targets, target_shape=(preds_height, preds_width))

            # Calculate metrics per class (start at index 0 (background) to cover cases where no mask exists)
            # ================================================================
            for cls_idx in range(0, num_classes):
                cls_target = (targets == cls_idx)
                cls_pred = preds[cls_idx]

                p = np.sum(cls_target)
                n = np.sum(~cls_target)
                if self.hard_threshold:
                    cls_pred_thresh = cls_pred > self.threshold
                    tp = np.sum(cls_pred_thresh[cls_target])  # sum intersection pixels - true positives
                    fp = np.sum(cls_pred_thresh[~cls_target])
                else:
                    tp = np.sum(cls_pred[cls_target])  # sum intersection pixels - true positives
                    fp = np.sum(cls_pred[~cls_target])  # sum pixels outside ground truth area - false positives
                tn = n - fp  # not used...
                fn = p - tp

                raw_metrics_per_class[cls_idx]['tp'] += tp
                raw_metrics_per_class[cls_idx]['fp'] += fp
                raw_metrics_per_class[cls_idx]['fn'] += fn

                # Collect global sums for the extra 'average' class
                raw_metrics_per_class[num_classes]['tp'] += tp
                raw_metrics_per_class[num_classes]['fp'] += fp
                raw_metrics_per_class[num_classes]['fn'] += fn

        if self._class_names is None:
            self._class_names = ['class_' + str(i) for i in range(1, num_classes)]
        self._class_names.append('micro_avg')

        result = {}
        for cls_idx in range(1, num_classes + 1):
            tp = raw_metrics_per_class[cls_idx]['tp']
            fn = raw_metrics_per_class[cls_idx]['fn']
            fp = raw_metrics_per_class[cls_idx]['fp']

            if (fn + tp + fp) > 0:
                iou = tp / (fn + tp + fp)
                dice = (2 * tp) / (fn + 2 * tp + fp)
            else:
                iou = np.nan
                dice = np.nan

            class_name = self._class_names[cls_idx - 1]
            result[class_name + '_iou'] = iou
            result[class_name + '_dice'] = dice

        return result



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

from typing import Optional

import torch
import pandas as pd
import numpy as np

from fuse.data.processor.processor_base import FuseProcessorBase


class FuseSkinGroundTruthProcessor(FuseProcessorBase):
    def __init__(self,
                 input_data: str,
                 train: Optional[bool] = True,
                 year: Optional[str] = '2016'):

        """
        Create Ground Truth labels
        :param input_data: path to labels
        :param train:      Optional, specify if we are in training phase
        :param year:       Optional, ISIC challenge year
        """

        self.input_data = input_data

        if year == '2016':
            input_df = pd.read_csv(input_data, names=['id', 'label'])
            if train:
                input_df.label = np.where(input_df.label == 'benign', 0, 1)
        else:  # year = 2017
            input_df = pd.read_csv(input_data, header=0, names=['id', 'label', 'other'])
        self.labels = dict(zip(input_df.id, input_df.label))

    def __call__(self,
                 sample_desc,
                 *args, **kwargs):

        result = {'tensor': torch.tensor(self.labels[sample_desc], dtype=torch.int64)}
        return result



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

from typing import Sequence, Dict, Tuple

import torch
import torch.nn as nn


from fuse.utils.utils_hierarchical_dict import FuseUtilsHierarchicalDict
import numpy as np
from fuse_examples.classification.prostate_x.head_1d_classifier import FuseHead1dClassifier


# 3x3 convolution
def conv3x3(in_channels, out_channels, stride=1):
    return nn.Conv3d(in_channels, out_channels, kernel_size=3,
                     stride=stride, padding=1, bias=False)



# Residual block
class ResidualBlock(torch.nn.Module):


    def __init__(self, in_channels, out_channels, stride=1, downsample=None):
        super(ResidualBlock, self).__init__()
        self.conv1 = conv3x3(in_channels, out_channels, stride)
        self.bn1 = nn.BatchNorm3d(out_channels)
        self.relu = nn.ReLU(inplace=True)
        self.conv2 = conv3x3(out_channels, out_channels)
        self.bn2 = nn.BatchNorm3d(out_channels)
        self.downsample = downsample



    def forward(self, x):
        residual = x
        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)
        out = self.conv2(out)
        out = self.bn2(out)
        res_mod = 'concat'
        if res_mod=='sum':
            if self.downsample:
                residual = self.downsample(x)
            out += residual
        elif res_mod=='concat':
            out = torch.cat((residual,out),1)
        out = self.relu(out)
        return out


# ResNet
class ResNet(torch.nn.Module):
    def __init__(self,
                 conv_inputs:Tuple[Tuple[str, int], ...] = (('data.input', 1),),
                 ch_num: None = None,
                 ) -> None:

        super(ResNet, self).__init__()
        block = ResidualBlock
        self.ch_num = ch_num
        layers = [1,1,1,1,1]
        out_features = [32,64,128,256,512]
        self.in_channels = 16
        in_features = [self.in_channels ,48,112,240,496]
        self.conv_inputs = conv_inputs
        if self.ch_num is None:
            self.conv = conv3x3(5, 16)
        else:
            self.conv = conv3x3(len(self.ch_num), 16)


        self.bn = nn.BatchNorm3d(16)
        self.relu = nn.ReLU(inplace=True)

        self.layer1 = self.make_layer(block, out_features[0], layers[0])
        self.in_channels = in_features[1]
        self.layer2 = self.make_layer(block, out_features[1], layers[1])
        self.in_channels = in_features[2]
        self.layer3 = self.make_layer(block, out_features[2], layers[2])
        self.in_channels = in_features[3]
        self.layer4 = self.make_layer(block, out_features[3], layers[3])
        self.in_channels = in_features[4]
        self.layer5 = self.make_layer(block, out_features[4], layers[4])
        self.max_pool = nn.MaxPool3d((2,2,2),stride=(2,2,2))
        self.conv_last = conv3x3(1008, 1008)
        self.bn_last = nn.BatchNorm3d(1008)

        self.dropout = nn.Dropout(0.5)
        self.fc1 = nn.Linear(1008, 2048)
        self.fc2 = nn.Linear(2048, 512)


    def make_layer(self, block, out_channels, blocks, stride=1):
        downsample = None
        if (stride != 1) or (self.in_channels != out_channels):
            downsample = nn.Sequential(
                conv3x3(self.in_channels, out_channels, stride=stride),
                nn.BatchNorm3d(out_channels))
        layers = []
        layers.append(block(self.in_channels, out_channels, stride, downsample))
        self.in_channels = out_channels
        for i in range(1, blocks):
            layers.append(block(out_channels, out_channels))
        return nn.Sequential(*layers)

    def forward(self,batch_dict):
        if len(self.conv_inputs)>1:
            tensors_list = [FuseUtilsHierarchicalDict.get(batch_dict, conv_input[0]).float() for conv_input in self.conv_inputs]
            max_tensor_dim = max([len(tmp_tensor.shape) for tmp_tensor in tensors_list])
            conv_input = torch.cat([tmp_tensor.unsqueeze_(0) if len(tmp_tensor.shape)<max_tensor_dim else tmp_tensor for tmp_tensor in tensors_list], 1)
        else:
            conv_input = torch.cat([FuseUtilsHierarchicalDict.get(batch_dict, conv_input[0]) for conv_input in self.conv_inputs], 1)

        if (self.ch_num is not None):
            if (conv_input.shape[0]>len(self.ch_num)):
                conv_input = conv_input[:,self.ch_num, :, :, :]


        if len(conv_input.shape)<4:
            conv_input = conv_input.unsqueeze_(0)

        out = self.conv(conv_input)
        out = self.bn(out)
        out = self.relu(out)
        out = self.max_pool(out)

        out = self.layer1(out)
        out = self.max_pool(out)
        out = self.layer2(out)
        out = self.max_pool(out)
        out = self.layer3(out)
        out = self.max_pool(out)
        out = self.layer4(out)
        out = self.max_pool(out)
        out = self.layer5(out)
        out = self.max_pool(out)

        out = self.conv_last(out)
        out = self.bn_last(out)
        out = self.relu(out)

        out = out.view(out.size(0), -1)
        out = self.fc1(out)
        out = self.relu(out)
        out = self.dropout(out)

        out = self.fc2(out)
        out = self.dropout(out)

        return out

class Fuse_model_3d_multichannel(torch.nn.Module):
    """
    Fuse model that classifing high resolution images
    Sequence:
        1. Starts with low resolution image to extract low resolution features and attention score per each image patch
        2. Extract high resolution features 'k' most significant patches (using the attention score)
        3. Use low resolution features, high resolution features and attention score to classify the image
    Input: high resolution tensor [BATCH_SIZE, 1, H, W]
    Output:
        TBD
    """

    def __init__(self,
                 conv_inputs: Tuple[Tuple[str, int], ...] = (('data.input', 1),),
                 backbone: ResNet = ResNet(),
                 heads: Sequence[torch.nn.Module] = (FuseHead1dClassifier(),),
                 ch_num = None,
                 ) -> None:
        """
        Default Fuse model - convolutional neural network with multiple heads
        :param conv_inputs:     batch_dict name for model input and its number of input channels
        :param backbone:        PyTorch backbone module - a convolutional neural network
        :param heads:           Sequence of head modules
        """
        super().__init__()

        self.conv_inputs = conv_inputs
        self.backbone = backbone
        self.heads = torch.nn.ModuleList(heads)
        self.add_module('heads', self.heads)
        self.ch_num = ch_num



    def forward(self,
                batch_dict: Dict) -> Dict:
        """
        Forward function of the model
        :param input: Tensor [BATCH_SIZE, 1, H, W]
        :return: classification scores - [BATCH_SIZE, num_classes]
        """

        features = self.backbone(batch_dict)
        FuseUtilsHierarchicalDict.set(batch_dict, 'model.backbone_features', features)
        FuseUtilsHierarchicalDict.set(batch_dict, 'model.backbone_features', features)

        for head in self.heads:
            batch_dict = head.forward(batch_dict)

        return batch_dict['model']

if __name__ == '__main__':

    import torch
    import os

    os.environ['CUDA_VISIBLE_DEVICES'] = "0, 1"

    DEVICE = 'cpu'  # 'cuda'
    DATAPARALLEL = False  # True
    num_features = 515
    model = Fuse_model_3d_multichannel(
        conv_inputs=(('data.input', 1),),
        backbone= ResNet(),
        heads=[
        FuseHead1dClassifier(head_name='ClinSig',
                                        conv_inputs=[('model.backbone_features', num_features)],
                                        post_concat_inputs=None,
                                        dropout_rate=0.25,
                                        shared_classifier_head=None,
                                        layers_description=None,
                                        num_classes=2),

        ]
    )

    model = model.to(DEVICE)
    if DATAPARALLEL:
        model = torch.nn.DataParallel(model)

    import gzip
    import pickle

    real_data_example = '/gpfs/haifa/projects/m/msieve_dev3/usr/Tal/fus_sessions/prostate/ProstateX_runs/V2/cache_prostate_x_size=74x74x13_spacing=0.5x0.5X3_fold0/0000000001.pkl.gz'
    with gzip.open(real_data_example, 'rb') as pickle_file:
        dummy_data = pickle.load(pickle_file)
        clinical_data = torch.rand([1]).to(DEVICE)
        clinical_data = clinical_data.unsqueeze(0)
        input = FuseUtilsHierarchicalDict.get(dummy_data, 'data.input')
        FuseUtilsHierarchicalDict.set(dummy_data,'data.input',input)
        zone = FuseUtilsHierarchicalDict.get(dummy_data, 'data.zone')
        zone2feature = {
                        'PZ':torch.tensor(np.array([0,0,0]),dtype=torch.float32).unsqueeze(0),
                        'TZ': torch.tensor(np.array([0,0,1]), dtype=torch.float32).unsqueeze(0),
                        'AS': torch.tensor(np.array([0,1,0]), dtype=torch.float32).unsqueeze(0),
                        'SV': torch.tensor(np.array([1,0,0]), dtype=torch.float32).unsqueeze(0),
                        }
        FuseUtilsHierarchicalDict.set(dummy_data, 'data.input', torch.cat((input.unsqueeze(0),input.unsqueeze(0)),dim=0))
        FuseUtilsHierarchicalDict.set(dummy_data, 'data.tensor_clinical', torch.cat((zone2feature[zone],zone2feature[zone]),dim=0))


    res = {}
    # manager = FuseManagerDefault()
    # manager.set_objects(net=model)
    # checkpoint =  '/gpfs/haifa/projects/m/msieve_dev3/usr/Tal/fus_sessions/multiclass_MG/malignant_multi_class_sentara_baptist_froedtert/exp_12_pretrain_normal-mal-benign_head/model_2/checkpoint_80_epoch.pth'
    # aa = manager.load_checkpoint(checkpoint)
    res['model'] = model.forward(dummy_data)
    print('Forward pass shape - head_0: ', end='')
    print(str(res['model']['logits']['head_1'].shape))

    print('\nForward pass shape - head_1: ', end='')
    print(str(res['model']['logits']['head_1'].shape))

    total_params = sum(p.numel() for p in model.parameters())
    if not DATAPARALLEL:
        backbone_params = sum(p.numel() for p in model._modules['backbone_clinical'].parameters())
        print('backbone_clinical params = %d' % backbone_params)
        print('Heads params = %d' % (total_params - backbone_params))

    print('\nTotal params = %d' % total_params)



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

from typing import Tuple
import os
import SimpleITK as sitk
import numpy as np
import torch
import logging

from fuse.data.processor.processor_base import FuseProcessorBase

from fuse_examples.classification.prostate_x.data_utils import FuseProstateXUtilsData
from fuse_examples.classification.prostate_x.preprocessing import PROSTATEX3DVolume



class FuseProstateXPatchProcessor(FuseProcessorBase):
    def __init__(self,
                 path_to_db: str,
                 data_path: str,
                 ktrans_data_path: str,
                 db_name: str,
                 fold_no : int,
                 lsn_shape: Tuple[int, int, int] = (16, 120, 120),
                 reference_inx: int = 0,
                 ):

        # store input parameters
        self.path_to_db = path_to_db
        self.data_path = data_path
        self.ktrans_data_path = ktrans_data_path
        self.lsn_shape = lsn_shape
        self.db_name = db_name
        self.fold_no=fold_no
        self.reference_inx = reference_inx


    def __call__(self,
                 sample_desc,
                 *args, **kwargs):
        """
        Return list of samples (lesions) giving a patient level descriptor
        :param sample_desc: (db_ver, set_type, patient_id)
        :return: list of lesions, see TorchClassificationAlgo.create_lesion_sample()
        """
        samples = []

        # decode descriptor
        db_ver, set_type, patient_id,pred_or_gt = sample_desc

        # get db - lesions
        db_full = FuseProstateXUtilsData.get_dataset(self.path_to_db,set_type,db_ver,self.db_name,self.fold_no)
        db = FuseProstateXUtilsData.get_lesions_prostate_x(db_full)

        # get patient
        patient = db[db['Patient ID'] == patient_id]

        # Print patient ID
        lgr = logging.getLogger('Fuse')
        lgr.info(f'patient={patient_id}', {'color': 'magenta'})

        # ---------------Read all images and masks, and then combine them---------------

        # Read all the modalities for a given ProstateX patient ID
        # There might be multiple directories (or not)
        self.prostate_data_path = self.data_path+'/PROSTATEx/'
        patient_directories = os.listdir(os.path.join(self.prostate_data_path, patient_id))
        if len(patient_directories) != 1:
            lgr.info(' - Warning: Multiple directories!')
        patient_directories = patient_directories[0]
        images_path = os.path.join(self.prostate_data_path, patient_id, patient_directories)

        # ------------------
        # initiate volume class
        vol3D = PROSTATEX3DVolume(patient_id=patient_id, imgs_path=images_path, ktrans_data_path=self.ktrans_data_path)
        # ------------------
        # list of volumes per sequence
        vol_list = vol3D.read_prostatex_sequences_patient()
        vol_ref = vol_list[self.reference_inx]
        #------------------
        # vol_4D is multichannel volume
        vol_4D = vol3D.preprocess_sequences(vol_list, reference_inx=self.reference_inx)


        #------------------
        # each row contains one lesion
        for index, row in patient.iterrows():

            pos_orig = np.array(np.fromstring(row.values[1], dtype=np.float32, sep=' '))
            pos_vol = np.array(vol_ref.TransformPhysicalPointToContinuousIndex(pos_orig.astype(np.float64)))

            vol_cropped, mask_cropped = vol3D.crop_lesion_vol(
                vol_4D, pos_vol,vol_ref ,center_slice=pos_vol[2],
                size=(self.lsn_shape[2], self.lsn_shape[1], self.lsn_shape[0]),
                spacing=(0.5, 0.5, 3))

            vol_cropped_tmp = sitk.GetArrayFromImage(vol_cropped)
            if len(vol_cropped_tmp.shape)<4:
                vol_cropped_tmp = vol_cropped_tmp[:,:,:,np.newaxis]
                vol = np.moveaxis(vol_cropped_tmp, 3, 0)
            else:
                vol = np.moveaxis(sitk.GetArrayFromImage(vol_cropped), 3, 0)

            if sum(sum(sum(sum(np.isnan(vol))))) > 0:
                input[np.isnan(input)] = 0

            mask = sitk.GetArrayFromImage(mask_cropped)
            vol_tensor = torch.from_numpy(vol).type(torch.FloatTensor)
            mask_tensor = torch.from_numpy(mask).unsqueeze(0).type(torch.FloatTensor)

            sample = {
                'patient_num': patient_id,
                'lesion_num': row['fid'],
                'input': vol_tensor,
                'input_lesion_mask': mask_tensor,
                'ggg': row['ggg'],
                'zone': row['zone'],
                'ClinSig': row['ClinSig'],

            }

            samples.append(sample)



        return samples


if __name__ == "__main__":
    path_to_db = '/gpfs/haifa/projects/m/msieve_dev3/usr/Tal/my_research/virtual_biopsy/prostate/experiments/V1/'
    dataset = 'prostate_x'
    if dataset=='prostate_x':
    # for ProstateX
        path_to_dataset = '/projects/msieve/MedicalSieve/PatientData/ProstateX/manifest-A3Y4AE4o5818678569166032044/'
        prostate_data_path = path_to_dataset
        Ktrain_data_path = path_to_dataset + '/ProstateXKtrains-train-fixed/'
        sample = ('18052021', 'train', 'ProstateX-0148', 'pred')

        a = FuseProstateXPatchProcessor(path_to_db,data_path=prostate_data_path,ktrans_data_path=Ktrain_data_path,db_name=dataset,fold_no=1,lsn_shape=(13, 74, 74),reference_inx=0)
        samples = a.__call__(sample)
        a = 1



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

from typing import Optional, Callable, List

import numpy as np

from fuse.metrics.metric_base import FuseMetricBase


class FuseMetricAccuracy(FuseMetricBase):
    """
    Basic Accuracy - Operation point is not used -
    the predicted class will simply be the one with the max probability
    """

    def __init__(self,
                 pred_name: str,
                 target_name: str,
                 class_names: Optional[List[str]] = None,
                 filter_func: Optional[Callable] = None,
                 use_sample_weights: Optional[bool] = False,
                 sample_weight_name: Optional[str] = None) -> None:
        """
        :param pred_name:           batch_dict key for predicted output (e.g., class probabilities after softmax)
        :param target_name:         batch_dict key for target (e.g., ground truth label)
        :param class_names:         Optional - name for each class otherwise the class index will be used.
        :param filter_func:         Optional - function that filters batch_dict. The function gets as input batch_dict and returns filtered batch_dict.
                                    When None (default), the entire batch dict is collected and processed.
        :param use_sample_weights:  Optional - when True (default is False), metrics are computed with sample weights.
                                    For this, use argument 'sample_weight_name' to define the batch_dict key for the sample weight
        :param sample_weight_name:  Optional, when use_sample_weights is True, user must specify the key for collecting sample weights
        """
        super().__init__(pred_name, target_name, filter_func=filter_func,
                         use_sample_weights=use_sample_weights, sample_weight_name=sample_weight_name)
        self._class_names = class_names

    def process(self) -> float:
        """
        :return: accuracy
        """
        if len(self.collected_data['pred']) == 0:
            return 'N/A'

        accurate = 0

        sample_weights = self.collected_data['sample_weight'] if self.use_sample_weights else np.ones(len(self.collected_data['target']))
        weight_sum = 0
        for target, pred, weight in zip(self.collected_data['target'], self.collected_data['pred'], sample_weights):
            pred = np.argmax(pred)
            if target == pred:
                accurate += weight
            weight_sum += weight

        accuracy = float(accurate) / weight_sum
        return accuracy


if __name__ == '__main__':
    data = {'preds': np.array([[0.8, 0.1, 0.1],
                               [0.5, 0.3, 0.2],
                               [0.6, 0.3, 0.1],
                               [0.6, 0.1, 0.3],
                               [0.7, 0.2, 0.1],
                               [0.3, 0.2, 0.5],
                               [0.1, 0.2, 0.7],
                               [0.2, 0.6, 0.2],
                               [0.3, 0.3, 0.4],
                               [0.7, 0.2, 0.1]]),
            'targets': np.array([0, 0, 0, 0, 0, 1, 1, 1, 2, 2]),
            'weights': np.array([0.03, 0.9, 0.05, 0.52, 0.23, 0.72, 0.13, 0.113, 0.84, 0.09])}

    metric = FuseMetricAccuracy(pred_name='preds', target_name='targets', use_sample_weights=True,
                           sample_weight_name='weights')
    metric.collect(data)
    print("use_sample_weights = True")
    res = metric.process()
    print(res)
    metric.reset()
    metric.collect(data)
    res = metric.process()
    print("use_sample_weights = True (after reset)")
    print(res)

    metric = FuseMetricAccuracy(pred_name='preds', target_name='targets', use_sample_weights=False,
                           sample_weight_name='weights')
    metric.collect(data)
    res = metric.process()
    print("use_sample_weights = False")
    print(res)


"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

"""
Cache to Memory
"""
from multiprocessing import Manager
from typing import Hashable, Any, List

from fuse.data.cache.cache_base import FuseCacheBase


class FuseCacheMemory(FuseCacheBase):
    """
    Cache to Memory
    """

    def __init__(self):
        super().__init__()

        self.reset()

    def __contains__(self, key: Hashable) -> bool:
        """
        See base class
        """
        return key in self._cache_dict

    def __getitem__(self, key: Hashable) -> Any:
        """
        See base class
        """
        return self._cache_dict.get(key, None)

    def __delitem__(self, key: Hashable) -> None:
        """
        See base class
        """
        if not self._cache_enable:
            raise Exception('First start caching using function start_caching()')

        item = self._cache_dict.pop(key, None)

    def __setitem__(self, key: Hashable, value: Any) -> None:
        """
        See base class
        """
        if not self._cache_enable:
            raise Exception('First start caching using function start_caching()')

        self._cache_dict[key] = value

    def save(self) -> None:
        """
        Not saving, moving back to  simple data structures
        """
        self._cache_enable = False
        self._cache_dict = dict(self._cache_dict)

    def exist(self) -> bool:
        """
        See base class
        """
        return len(self._cache_dict) > 0

    def reset(self) -> None:
        """
        See base class
        """
        self._cache_dict = {}

    def get_all_keys(self, include_none: bool = False) -> List[Hashable]:
        """
        See base class
        """
        if include_none:
            return list(self._cache_dict.keys())
        else:
            return [key for key, value in self._cache_dict.items() if value is not None]

    def start_caching(self, manager: Manager) -> None:
        """
        Moving to multiprocessing data structures
        """
        self._cache_enable = True
        # if manager is  None assume that the it's not multiprocessing caching
        if manager is not None:
            self._cache_dict = manager.dict(self._cache_dict)



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

from typing import Dict
from fuse.managers.callbacks.callback_base import FuseCallback
from fuse.managers.manager_state import FuseManagerState

class FuseCallbackOptClosure(FuseCallback):
    """
    Use this callback if an optimizer requires closure argument
    """
    def on_train_begin(self, state: FuseManagerState):
        self.virtual_batch = []
        self.state = state
        self.state.opt_closure = self.opt_closure

    def on_virtual_batch_begin(self, mode: str, virtual_batch: int):
        self.virtual_batch = []

    def on_batch_end(self, mode: str, batch: int, batch_dict: Dict = None):
        self.virtual_batch.append(batch_dict)

    def opt_closure(self) -> None:
        for batch in self.virtual_batch:
            # forward pass
            batch['model'] = self.state.net(batch)

            # compute loss
            total_loss: torch.Tensor = 0
            for loss_name, loss_function in self.state.losses.items():
                current_loss_result = loss_function(batch)
                # sum all losses for backward
                total_loss += current_loss_result
            total_loss.backward()


"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

"""
Fuse Files Utils
"""
import errno
import logging
import os
import shutil
from typing import Iterable

from fuse.utils.utils_misc import FuseUtilsMisc


class FuseUtilsFile:
    @staticmethod
    def create_dir(dir_path: str, error_if_exist: bool = False) -> None:
        """
        Create dir
        :param dir_path: path to dir. either relative or full
        :param error_if_exist: if false will ignore already exist error
        :return: None
        """
        if not os.path.isdir(dir_path):
            try:
                os.makedirs(dir_path)
            except OSError as e:
                if e.errno != errno.EEXIST or error_if_exist:
                    raise

    @staticmethod
    def remove_dir_content(dir_path: str, ignore_files: Iterable[str] = tuple(), force_reset: bool = False) -> None:
        """
        Remove the content of dir_path ignoring the files under ignore_files.
        If force_reset is False, prompts the user for approval before the deletion.

        :param dir_path: path to dir. either relative or full
        :param ignore_files: list of files to ignore (don't delete them)
        :param force_reset: when False (default), asks the user for approval when deleting content.
            Else, delete without prompting.
        :return: None
        """
        lgr = logging.getLogger('Fuse')

        # if no content - do nothing
        files = os.listdir(dir_path)
        files = [file for file in files if file not in ignore_files]
        num_files = len(files)
        if num_files > 0:
            # prompt the user for approval
            force_remove = True if force_reset else FuseUtilsMisc.query_yes_no(
                question=f'Folder {dir_path} contains {num_files} files. Delete anyway?')
            # iterate over all the content and delete it
            failure = False
            if force_remove:
                for filename in files:
                    file_path = os.path.join(dir_path, filename)
                    try:
                        if os.path.isfile(file_path) or os.path.islink(file_path):
                            os.unlink(file_path)
                        elif os.path.isdir(file_path):
                            shutil.rmtree(file_path)
                    except Exception as e:
                        lgr.error('Failed to delete %s. Reason: %s' % (file_path, e))
                        failure = True

            # in case remove wasn't approved or failed to delete some of the files
            if not force_remove or failure:
                msg = f'Folder {dir_path} is already used, please remove it and rerun the program'
                lgr.error(msg)
                raise Exception(msg)

    @staticmethod
    def create_or_reset_dir(dir_path: str, ignore_files: Iterable[str] = tuple(), force_reset: bool = False) -> None:
        """
        Create dir or reset it if already exists
        :param dir_path: path to dir. either relative or full
        :param ignore_files: list of files to ignore (don't delete them)
        :return: None
        """
        FuseUtilsFile.create_dir(dir_path)
        FuseUtilsFile.remove_dir_content(dir_path, ignore_files, force_reset)
